from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.StorageManagementSystem.DB.StorageManagementDB import THROTTLING_STEPS, THROTTLING_TIME

import re

AGENT_NAME = "StorageManagement/StageRequestAgent"


class StageRequestAgent(AgentModule):
    def initialize(self):
        self.stagerClient = StorageManagerClient()
        # self.storageDB = StorageManagementDB()
        # pin lifetime = 1 day
        self.pinLifetime = self.am_getOption("PinLifetime", THROTTLING_TIME)

        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")

        return S_OK()

    def execute(self):

        # Get the current submitted stage space and the amount of pinned space for each storage element
        res = self.getStorageUsage()
        if not res["OK"]:
            return res

        return self.submitStageRequests()

    def getStorageUsage(self):
        """Fill the current Status of the SE Caches from the DB"""
        self.storageElementCache = {}

        res = self.stagerClient.getSubmittedStagePins()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest.getStorageUsage: Failed to obtain submitted requests from StorageManagementDB.",
                res["Message"],
            )
            return res
        self.storageElementUsage = res["Value"]
        if self.storageElementUsage:
            gLogger.info("StageRequest.getStorageUsage: Active stage/pin requests found at the following sites:")
            for storageElement in sorted(self.storageElementUsage.keys()):
                seDict = self.storageElementUsage[storageElement]
                # Convert to GB for printout
                seDict["TotalSize"] = seDict["TotalSize"] / (1000 * 1000 * 1000.0)
                gLogger.info(
                    "StageRequest.getStorageUsage: %s: %s replicas with a size of %.3f GB."
                    % (storageElement.ljust(15), str(seDict["Replicas"]).rjust(6), seDict["TotalSize"])
                )
        if not self.storageElementUsage:
            gLogger.info("StageRequest.getStorageUsage: No active stage/pin requests found.")

        return S_OK()

    def submitStageRequests(self):
        """This manages the following transitions of the Replicas
        * Waiting -> Offline (if the file is not found Cached)
        * Waiting -> StageSubmitted (if the file is found Cached)
        * Offline -> StageSubmitted (if there are not more Waiting replicas)
        """
        # Retry Replicas that have not been Staged in a previous attempt
        res = self._getMissingReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest.submitStageRequests: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        seReplicas = res["Value"]["SEReplicas"]
        allReplicaInfo = res["Value"]["AllReplicaInfo"]

        if seReplicas:
            gLogger.info("StageRequest.submitStageRequests: Completing partially Staged Tasks")
        for storageElement, seReplicaIDs in seReplicas.items():
            gLogger.debug("Staging at %s:" % storageElement, seReplicaIDs)
            self._issuePrestageRequests(storageElement, seReplicaIDs, allReplicaInfo)

        # Check Waiting Replicas and select those found Online and all other Replicas from the same Tasks
        res = self._getOnlineReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest.submitStageRequests: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        seReplicas = res["Value"]["SEReplicas"]
        allReplicaInfo = res["Value"]["AllReplicaInfo"]

        # Check Offline Replicas that fit in the Cache and all other Replicas from the same Tasks
        res = self._getOfflineReplicas()

        if not res["OK"]:
            gLogger.fatal(
                "StageRequest.submitStageRequests: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res

        # Merge info from both results
        for storageElement, seReplicaIDs in res["Value"]["SEReplicas"].items():
            seReplicas.setdefault(storageElement, []).extend(seReplicaIDs)
        allReplicaInfo.update(res["Value"]["AllReplicaInfo"])

        gLogger.info("StageRequest.submitStageRequests: Obtained %s replicas for staging." % len(allReplicaInfo))
        for storageElement, seReplicaIDs in seReplicas.items():
            gLogger.debug("Staging at %s:" % storageElement, seReplicaIDs)
            self._issuePrestageRequests(storageElement, seReplicaIDs, allReplicaInfo)
        return S_OK()

    def _getMissingReplicas(self):
        """This recovers Replicas that were not Staged on a previous attempt (the stage request failed or timed out),
        while other Replicas of the same task are already Staged. If left behind they can produce a deadlock.
        All SEs are considered, even if their Cache is full
        """
        # Get Replicas that are in Staged/StageSubmitted
        gLogger.info("StageRequest._getMissingReplicas: Checking Staged Replicas")

        res = self.__getStagedReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest._getMissingReplicas: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        seReplicas = {}

        allReplicaInfo = res["Value"]["AllReplicaInfo"]
        replicasToStage = []
        for seReplicaIDs in res["Value"]["SEReplicas"].values():
            # Consider all SEs
            replicasToStage += seReplicaIDs

        # Get Replicas from the same Tasks as those selected
        res = self.__addAssociatedReplicas(replicasToStage, seReplicas, allReplicaInfo)
        if not res["OK"]:
            gLogger.fatal("StageRequest._getMissingReplicas: Failed to get associated Replicas.", res["Message"])

        return res

    def _getOnlineReplicas(self):
        """This manages the transition
        * Waiting -> Offline (if the file is not found Cached)
        and returns the list of Cached Replicas for which the pin time has to be extended
        SEs for which the cache is currently full are not considered
        """
        # Get all Replicas in Waiting Status associated to Staging Tasks
        gLogger.verbose("StageRequest._getOnlineReplicas: Checking Online Replicas to be handled")

        res = self.__getWaitingReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest._getOnlineReplicas: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        seReplicas = {}
        allReplicaInfo = res["Value"]["AllReplicaInfo"]
        if not len(allReplicaInfo):
            gLogger.info("StageRequest._getOnlineReplicas: There were no Waiting replicas found")
            return res
        gLogger.info("StageRequest._getOnlineReplicas: Obtained %s replicas Waiting for staging." % len(allReplicaInfo))
        replicasToStage = []
        for storageElement, seReplicaIDs in res["Value"]["SEReplicas"].items():
            if not self.__usage(storageElement) < self.__cache(storageElement):
                gLogger.info(
                    "StageRequest._getOnlineReplicas: Skipping %s, current usage above limit ( %s GB )"
                    % (storageElement, self.__cache(storageElement))
                )
                # Do not consider those SE that have the Cache full
                continue
            # Check if the Replica Metadata is OK and find out if they are Online or Offline
            res = self.__checkIntegrity(storageElement, seReplicaIDs, allReplicaInfo)
            if not res["OK"]:
                gLogger.error(
                    "StageRequest._getOnlineReplicas: Failed to check Replica Metadata",
                    "(%s): %s" % (storageElement, res["Message"]),
                )
            else:
                # keep only Online Replicas
                seReplicas[storageElement] = res["Value"]["Online"]
                replicasToStage += res["Value"]["Online"]

        # Get Replicas from the same Tasks as those selected
        res = self.__addAssociatedReplicas(replicasToStage, seReplicas, allReplicaInfo)
        if not res["OK"]:
            gLogger.fatal("StageRequest._getOnlineReplicas: Failed to get associated Replicas.", res["Message"])

        return res

    def _getOfflineReplicas(self):
        """This checks Replicas in Offline status
        and returns the list of Replicas to be Staged
        SEs for which the cache is currently full are not considered
        """
        # Get all Replicas in Waiting Status associated to Staging Tasks
        gLogger.verbose("StageRequest._getOfflineReplicas: Checking Offline Replicas to be handled")

        res = self.__getOfflineReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "StageRequest._getOfflineReplicas: Failed to get replicas from StorageManagementDB.", res["Message"]
            )
            return res
        seReplicas = {}
        allReplicaInfo = res["Value"]["AllReplicaInfo"]
        if not len(allReplicaInfo):
            gLogger.info("StageRequest._getOfflineReplicas: There were no Offline replicas found")
            return res
        gLogger.info(
            "StageRequest._getOfflineReplicas: Obtained %s replicas Offline for staging." % len(allReplicaInfo)
        )
        replicasToStage = []
        for storageElement, seReplicaIDs in res["Value"]["SEReplicas"].items():
            if not self.__usage(storageElement) < self.__cache(storageElement):
                gLogger.info(
                    "StageRequest._getOfflineReplicas: Skipping %s, current usage above limit ( %s GB )"
                    % (storageElement, self.__cache(storageElement))
                )
                # Do not consider those SE that have the Cache full
                continue
            seReplicas[storageElement] = []
            for replicaID in sorted(seReplicaIDs):
                seReplicas[storageElement].append(replicaID)
                replicasToStage.append(replicaID)
                self.__add(storageElement, allReplicaInfo[replicaID]["Size"])
                if not self.__usage(storageElement) < self.__cache(storageElement):
                    # Stop adding Replicas when the cache is full
                    break

        # Get Replicas from the same Tasks as those selected
        res = self.__addAssociatedReplicas(replicasToStage, seReplicas, allReplicaInfo)
        if not res["OK"]:
            gLogger.fatal("StageRequest._getOfflineReplicas: Failed to get associated Replicas.", res["Message"])

        return res

    def __usage(self, storageElement):
        """Retrieve current usage of SE"""
        # Set it if not yet done
        self.storageElementUsage.setdefault(storageElement, {"TotalSize": 0.0})
        return self.storageElementUsage[storageElement]["TotalSize"]

    def __cache(self, storageElement):
        """Retrieve cache size for SE"""
        if storageElement not in self.storageElementCache:
            diskCacheTB = float(StorageElement(storageElement).options.get("DiskCacheTB", 1.0))
            self.storageElementCache[storageElement] = diskCacheTB * 1000.0 / THROTTLING_STEPS
        return self.storageElementCache[storageElement]

    def __add(self, storageElement, size):
        """Add size (in bytes) to current usage of storageElement (in GB)"""
        self.storageElementUsage.setdefault(storageElement, {"TotalSize": 0.0})
        size /= 1000.0 * 1000.0 * 1000.0
        self.storageElementUsage[storageElement]["TotalSize"] += size
        return size

    def _issuePrestageRequests(self, storageElement, seReplicaIDs, allReplicaInfo):
        """Make the request to the SE and update the DB"""
        # Since we are in a give SE, the lfn is a unique key
        lfnRepIDs = {}
        for replicaID in seReplicaIDs:
            lfn = allReplicaInfo[replicaID]["LFN"]
            lfnRepIDs[lfn] = replicaID

        # Now issue the prestage requests for the remaining replicas
        stageRequestMetadata = {}
        updatedLfnIDs = []
        if lfnRepIDs:
            gLogger.info(
                "StageRequest._issuePrestageRequests: Submitting %s stage requests for %s."
                % (len(lfnRepIDs), storageElement)
            )
            res = StorageElement(storageElement).prestageFile(lfnRepIDs, lifetime=self.pinLifetime)
            gLogger.debug("StageRequest._issuePrestageRequests: StorageElement.prestageStorageFile: res=", res)
            # Daniela: fishy result from ReplicaManager!!! Should NOT return OK
            # res= {'OK': True, 'Value': {'Successful': {}, 'Failed': {'srm://srm-lhcb.cern.ch/castor/cern.ch/grid/lhcb/data/2010/RAW/EXPRESS/LHCb/COLLISION10/71476/071476_0000000241.raw': ' SRM2Storage.__gfal_exec: Failed to perform gfal_prestage.[SE][BringOnline][SRM_INVALID_REQUEST] httpg://srm-lhcb.cern.ch:8443/srm/managerv2: User not able to access specified space token\n'}}}
            # res= {'OK': True, 'Value': {'Successful': {'srm://gridka-dCache.fzk.de/pnfs/gridka.de/lhcb/data/2009/RAW/FULL/LHCb/COLLISION09/63495/063495_0000000001.raw': '-2083846379'}, 'Failed': {}}}

            if not res["OK"]:
                gLogger.error(
                    "StageRequest._issuePrestageRequests: Completely failed to submit stage requests for replicas.",
                    res["Message"],
                )
            else:
                for lfn, requestID in res["Value"]["Successful"].items():
                    stageRequestMetadata.setdefault(requestID, []).append(lfnRepIDs[lfn])
                    updatedLfnIDs.append(lfnRepIDs[lfn])
        if stageRequestMetadata:
            gLogger.info(
                "StageRequest._issuePrestageRequests: %s stage request metadata to be updated."
                % len(stageRequestMetadata)
            )
            res = self.stagerClient.insertStageRequest(stageRequestMetadata, self.pinLifetime)
            if not res["OK"]:
                gLogger.error(
                    "StageRequest._issuePrestageRequests: Failed to insert stage request metadata.", res["Message"]
                )
                return res
            res = self.stagerClient.updateReplicaStatus(updatedLfnIDs, "StageSubmitted")
            if not res["OK"]:
                gLogger.error("StageRequest._issuePrestageRequests: Failed to insert replica status.", res["Message"])
        return

    def __sortBySE(self, replicaDict):

        seReplicas = {}
        replicaIDs = {}
        for replicaID, info in replicaDict.items():
            lfn = info["LFN"]
            storageElement = info["SE"]
            size = info["Size"]
            pfn = info["PFN"]
            replicaIDs[replicaID] = {"LFN": lfn, "PFN": pfn, "Size": size, "StorageElement": storageElement}
            seReplicas.setdefault(storageElement, []).append(replicaID)
        return S_OK({"SEReplicas": seReplicas, "AllReplicaInfo": replicaIDs})

    def __getStagedReplicas(self):
        """This obtains the Staged replicas from the Replicas table and for each LFN the requested storage element"""
        # First obtain the Waiting replicas from the Replicas table
        res = self.stagerClient.getStagedReplicas()
        if not res["OK"]:
            gLogger.error(
                "StageRequest.__getStagedReplicas: Failed to get replicas with Waiting status.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.debug("StageRequest.__getStagedReplicas: No Waiting replicas found to process.")
        else:
            gLogger.debug(
                "StageRequest.__getStagedReplicas: Obtained %s Waiting replicas(s) to process." % len(res["Value"])
            )

        return self.__sortBySE(res["Value"])

    def __getWaitingReplicas(self):
        """This obtains the Waiting replicas from the Replicas table and for each LFN the requested storage element"""
        # First obtain the Waiting replicas from the Replicas table
        res = self.stagerClient.getWaitingReplicas()
        if not res["OK"]:
            gLogger.error(
                "StageRequest.__getWaitingReplicas: Failed to get replicas with Waiting status.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.debug("StageRequest.__getWaitingReplicas: No Waiting replicas found to process.")
        else:
            gLogger.debug(
                "StageRequest.__getWaitingReplicas: Obtained %s Waiting replicas(s) to process." % len(res["Value"])
            )

        return self.__sortBySE(res["Value"])

    def __getOfflineReplicas(self):
        """This obtains the Offline replicas from the Replicas table and for each LFN the requested storage element"""
        # First obtain the Waiting replicas from the Replicas table
        res = self.stagerClient.getOfflineReplicas()
        if not res["OK"]:
            gLogger.error(
                "StageRequest.__getOfflineReplicas: Failed to get replicas with Waiting status.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.debug("StageRequest.__getOfflineReplicas: No Waiting replicas found to process.")
        else:
            gLogger.debug(
                "StageRequest.__getOfflineReplicas: Obtained %s Waiting replicas(s) to process." % len(res["Value"])
            )

        return self.__sortBySE(res["Value"])

    def __addAssociatedReplicas(self, replicasToStage, seReplicas, allReplicaInfo):
        """Retrieve the list of Replicas that belong to the same Tasks as the provided list"""
        res = self.stagerClient.getAssociatedReplicas(replicasToStage)
        if not res["OK"]:
            gLogger.fatal("StageRequest.__addAssociatedReplicas: Failed to get associated Replicas.", res["Message"])
            return res
        addReplicas = {"Offline": {}, "Waiting": {}}
        replicaIDs = {}
        for replicaID, info in res["Value"].items():
            lfn = info["LFN"]
            storageElement = info["SE"]
            size = info["Size"]
            pfn = info["PFN"]
            status = info["Status"]
            if status in ["Waiting", "Offline"]:
                replicaIDs[replicaID] = {"LFN": lfn, "PFN": pfn, "Size": size, "StorageElement": storageElement}
                addReplicas[status].setdefault(storageElement, []).append(replicaID)

        waitingReplicas = addReplicas["Waiting"]
        offlineReplicas = addReplicas["Offline"]
        newReplicaInfo = replicaIDs
        allReplicaInfo.update(newReplicaInfo)

        # First handle Waiting Replicas for which metadata is to be checked
        for storageElement, seReplicaIDs in waitingReplicas.items():
            for replicaID in list(seReplicaIDs):
                if replicaID in replicasToStage:
                    seReplicaIDs.remove(replicaID)
            res = self.__checkIntegrity(storageElement, seReplicaIDs, allReplicaInfo)
            if not res["OK"]:
                gLogger.error(
                    "StageRequest.__addAssociatedReplicas: Failed to check Replica Metadata",
                    "(%s): %s" % (storageElement, res["Message"]),
                )
            else:
                # keep all Replicas (Online and Offline)
                seReplicas.setdefault(storageElement, []).extend(res["Value"]["Online"])
                replicasToStage.extend(res["Value"]["Online"])
                seReplicas[storageElement].extend(res["Value"]["Offline"])
                replicasToStage.extend(res["Value"]["Offline"])

        # Then handle Offline Replicas for which metadata is already checked
        for storageElement, seReplicaIDs in offlineReplicas.items():
            for replicaID in sorted(seReplicaIDs):
                if replicaID in replicasToStage:
                    seReplicaIDs.remove(replicaID)
            seReplicas.setdefault(storageElement, []).extend(seReplicaIDs)
            replicasToStage.extend(seReplicaIDs)

        for replicaID in list(allReplicaInfo):
            if replicaID not in replicasToStage:
                del allReplicaInfo[replicaID]

        totalSize = 0
        for storageElement in sorted(seReplicas.keys()):
            replicaIDs = seReplicas[storageElement]
            size = 0
            for replicaID in replicaIDs:
                size += self.__add(storageElement, allReplicaInfo[replicaID]["Size"])

            gLogger.info(
                "StageRequest.__addAssociatedReplicas:  Considering %s GB to be staged at %s" % (size, storageElement)
            )
            totalSize += size

        gLogger.info("StageRequest.__addAssociatedReplicas: Obtained %s GB for staging." % totalSize)

        return S_OK({"SEReplicas": seReplicas, "AllReplicaInfo": allReplicaInfo})

    def __checkIntegrity(self, storageElement, seReplicaIDs, allReplicaInfo):
        """Check the integrity of the files to ensure they are available
        Updates status of Offline Replicas for a later pass
        Return list of Online replicas to be Stage
        """
        if not seReplicaIDs:
            return S_OK({"Online": [], "Offline": []})

        # Since we are with a given SE, the LFN is a unique key
        lfnRepIDs = {}
        for replicaID in seReplicaIDs:
            lfn = allReplicaInfo[replicaID]["LFN"]
            lfnRepIDs[lfn] = replicaID

        gLogger.info(
            "StageRequest.__checkIntegrity: Checking the integrity of %s replicas at %s."
            % (len(lfnRepIDs), storageElement)
        )
        res = StorageElement(storageElement).getFileMetadata(lfnRepIDs)
        if not res["OK"]:
            gLogger.error(
                "StageRequest.__checkIntegrity: Completely failed to obtain metadata for replicas.", res["Message"]
            )
            return res

        terminalReplicaIDs = {}
        onlineReplicaIDs = []
        offlineReplicaIDs = []
        for lfn, metadata in res["Value"]["Successful"].items():

            if metadata["Size"] != allReplicaInfo[lfnRepIDs[lfn]]["Size"]:
                gLogger.error("StageRequest.__checkIntegrity: LFN StorageElement size does not match FileCatalog", lfn)
                terminalReplicaIDs[lfnRepIDs[lfn]] = "LFN StorageElement size does not match FileCatalog"
                lfnRepIDs.pop(lfn)
            elif metadata.get("Lost", False):
                gLogger.error("StageRequest.__checkIntegrity: LFN has been Lost by the StorageElement", lfn)
                terminalReplicaIDs[lfnRepIDs[lfn]] = "LFN has been Lost by the StorageElement"
                lfnRepIDs.pop(lfn)
            elif metadata.get("Unavailable", False):
                gLogger.error("StageRequest.__checkIntegrity: LFN is declared Unavailable by the StorageElement", lfn)
                terminalReplicaIDs[lfnRepIDs[lfn]] = "LFN is declared Unavailable by the StorageElement"
                lfnRepIDs.pop(lfn)
            elif metadata.get("Cached", metadata["Accessible"]):
                gLogger.verbose("StageRequest.__checkIntegrity: Cache hit for file.")
                onlineReplicaIDs.append(lfnRepIDs[lfn])
            else:
                offlineReplicaIDs.append(lfnRepIDs[lfn])

        for lfn, reason in res["Value"]["Failed"].items():
            if re.search("File does not exist", reason):
                gLogger.error("StageRequest.__checkIntegrity: LFN does not exist in the StorageElement", lfn)
                terminalReplicaIDs[lfnRepIDs[lfn]] = "LFN does not exist in the StorageElement"
            lfnRepIDs.pop(lfn)

        # Update the states of the replicas in the database #TODO Sent status to integrity DB
        if terminalReplicaIDs:
            gLogger.info("StageRequest.__checkIntegrity: %s replicas are terminally failed." % len(terminalReplicaIDs))
            res = self.stagerClient.updateReplicaFailure(terminalReplicaIDs)
            if not res["OK"]:
                gLogger.error("StageRequest.__checkIntegrity: Failed to update replica failures.", res["Message"])
        if onlineReplicaIDs:
            gLogger.info("StageRequest.__checkIntegrity: %s replicas found Online." % len(onlineReplicaIDs))
        if offlineReplicaIDs:
            gLogger.info("StageRequest.__checkIntegrity: %s replicas found Offline." % len(offlineReplicaIDs))
            res = self.stagerClient.updateReplicaStatus(offlineReplicaIDs, "Offline")
        return S_OK({"Online": onlineReplicaIDs, "Offline": offlineReplicaIDs})



""" GraphUtilities is a a collection of utility functions and classes used
    in the DIRAC Graphs package.

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import six
import os
import time
import datetime
import calendar
import math
import pytz
import numpy

from matplotlib.pyplot import rcParams
from matplotlib.ticker import ScalarFormatter
from matplotlib.dates import (
    AutoDateLocator,
    AutoDateFormatter,
    DateFormatter,
    RRuleLocator,
    rrulewrapper,
    HOURLY,
    MINUTELY,
    SECONDLY,
    YEARLY,
    MONTHLY,
    DAILY,
)
from dateutil.relativedelta import relativedelta

# This is a hack to workaround the use of float(ScalarFormatter.__call__(...))
rcParams["axes.unicode_minus"] = False


def evalPrefs(*args, **kw):
    """Interpret arguments as preferencies dictionaries or key-value pairs. The overriding order
    is right most - most important one. Returns a single dictionary of preferencies
    """

    prefs = {}
    for pDict in list(args) + [kw]:
        if isinstance(pDict, dict):
            for key in pDict:
                if key == "metadata":
                    for mkey in pDict[key]:
                        prefs[mkey] = pDict[key][mkey]
                else:
                    prefs[key] = pDict[key]

    return prefs


def pixelToPoint(size, dpi):
    """Convert size expressed in pixels into points for a given dpi resolution"""

    return float(size) * 100.0 / float(dpi)


datestrings = ["%x %X", "%x", "%Y-%m-%d %H:%M:%S"]


def convert_to_datetime(dstring):
    orig_string = str(dstring)
    try:
        if isinstance(dstring, datetime.datetime):
            results = dstring
        else:
            results = eval(str(dstring), {"__builtins__": None, "time": time, "math": math}, {})
        if isinstance(results, six.integer_types + (float,)):
            results = datetime.datetime.fromtimestamp(int(results))
        elif isinstance(results, datetime.datetime):
            pass
        else:
            raise ValueError("Unknown datetime type!")
    except Exception as e:
        t = None
        for dateformat in datestrings:
            try:
                t = time.strptime(dstring, dateformat)
                timestamp = calendar.timegm(t)  # -time.timezone
                results = datetime.datetime.fromtimestamp(timestamp)
                break
            except Exception:
                pass
        if t is None:
            try:
                dstring = dstring.split(".", 1)[0]
                t = time.strptime(dstring, dateformat)
                timestamp = time.mktime(t)  # -time.timezone
                results = datetime.datetime.fromtimestamp(timestamp)
            except Exception:
                raise ValueError(
                    "Unable to create time from string!\nExpecting "
                    "format of: '12/06/06 12:54:67'\nRecieved:%s" % orig_string
                )
    return results


def to_timestamp(val):

    try:
        v = float(val)
        if v > 1000000000 and v < 1900000000:
            return v
    except Exception:
        pass

    val = convert_to_datetime(val)
    # return calendar.timegm( val.timetuple() )
    return time.mktime(val.timetuple())


# If the graph has more than `hour_switch` minutes, we print
# out hours in the subtitle.
hour_switch = 7

# If the graph has more than `day_switch` hours, we print
# out days in the subtitle.
day_switch = 7

# If the graph has more than `week_switch` days, we print
# out the weeks in the subtitle.
week_switch = 7


def add_time_to_title(begin, end, metadata={}):
    """Given a title and two times, adds the time info to the title.
    Example results::

       "Number of Attempted Transfers
       (24 Hours from 4:45 12-14-2006 to 5:56 12-15-2006)"

    There are two important pieces to the subtitle we add - the duration
    (i.e., '48 Hours') and the time interval (i.e., 11:00 07-02-2007 to
    11:00 07-04-2007).

    We attempt to make the duration match the size of the span (for a bar
    graph, this would be the width of the individual bar) in order for it
    to make the most sense.  The formatting of the time interval is based
    upon how much real time there is from the beginning to the end.

    We made the distinction because some would want to show graphs
    representing 168 Hours, but needed the format to show the date as
    well as the time.
    """
    if "span" in metadata:
        interval = metadata["span"]
    else:
        interval = time_interval(begin, end)
    formatting_interval = time_interval(begin, end)
    if formatting_interval == 600:
        format_str = "%H:%M:%S"
    elif formatting_interval == 3600:
        format_str = "%Y-%m-%d %H:%M"
    elif formatting_interval == 86400:
        format_str = "%Y-%m-%d"
    elif formatting_interval == 86400 * 7:
        format_str = "Week %U of %Y"

    if interval < 600:
        format_name = "Seconds"
        time_slice = 1
    elif interval < 3600 and interval >= 600:
        format_name = "Minutes"
        time_slice = 60
    elif interval >= 3600 and interval < 86400:
        format_name = "Hours"
        time_slice = 3600
    elif interval >= 86400 and interval < 86400 * 7:
        format_name = "Days"
        time_slice = 86400
    elif interval >= 86400 * 7:
        format_name = "Weeks"
        time_slice = 86400 * 7
    else:
        format_str = "%x %X"
        format_name = "Seconds"
        time_slice = 1

    begin_tuple = time.localtime(begin)
    end_tuple = time.localtime(end)
    added_title = "%i %s from " % (int((end - begin) / time_slice), format_name)
    added_title += time.strftime("%s to" % format_str, begin_tuple)
    if time_slice < 86400:
        add_utc = " UTC"
    else:
        add_utc = ""
    added_title += time.strftime(" %s%s" % (format_str, add_utc), end_tuple)
    return added_title


def time_interval(begin, end):
    """
    Determine the appropriate time interval based upon the length of
    time as indicated by the `starttime` and `endtime` keywords.
    """

    if end - begin < 600 * hour_switch:
        return 600
    if end - begin < 86400 * day_switch:
        return 3600
    elif end - begin < 86400 * 7 * week_switch:
        return 86400
    else:
        return 86400 * 7


def comma_format(x_orig):
    x = float(x_orig)
    if x >= 1000:
        after_comma = x % 1000
        before_comma = int(int(x) / 1000)
        return "%s,%03g" % (comma_format(before_comma), after_comma)
    else:
        return str(x_orig)


class PrettyScalarFormatter(ScalarFormatter):
    def __init__(self, *args, **kwargs):
        super(PrettyScalarFormatter, self).__init__(*args, **kwargs)
        self.set_powerlimits([-7, 9])
        self._useLocale = True

    def __call__(self, x, pos=None):
        val = super(PrettyScalarFormatter, self).__call__(x, pos=pos)
        if self.offset:
            return val
        else:
            return comma_format(val)


class PrettyDateFormatter(AutoDateFormatter):
    """This class provides a formatter which conforms to the
    desired date formats for the Phedex system.
    """

    def __init__(self, locator):
        """Format dates according to the Phedex system"""
        tz = pytz.timezone("UTC")
        AutoDateFormatter.__init__(self, locator, tz=tz)

    def __call__(self, x, pos=0):
        scale = float(self._locator._get_unit())
        if scale == 365.0:
            self._formatter = DateFormatter("%Y", self._tz)
        elif scale == 30.0:
            self._formatter = DateFormatter("%b %Y", self._tz)
        elif (scale >= 1.0) and (scale <= 7.0):
            self._formatter = DateFormatter("%Y-%m-%d", self._tz)
        elif scale == (1.0 / 24.0):
            self._formatter = DateFormatter("%H:%M", self._tz)
        elif scale == (1.0 / (24 * 60)):
            self._formatter = DateFormatter("%H:%M", self._tz)
        elif scale == (1.0 / (24 * 3600)):
            self._formatter = DateFormatter("%H:%M:%S", self._tz)
        else:
            self._formatter = DateFormatter("%b %d %Y %H:%M:%S", self._tz)

        return self._formatter(x, pos)


class PrettyDateLocator(AutoDateLocator):
    def get_locator(self, dmin, dmax):
        "pick the best locator based on a distance"

        delta = relativedelta(dmax, dmin)
        numYears = delta.years * 1.0
        numMonths = (numYears * 12.0) + delta.months
        numDays = (numMonths * 31.0) + delta.days
        numHours = (numDays * 24.0) + delta.hours
        numMinutes = (numHours * 60.0) + delta.minutes
        numSeconds = (numMinutes * 60.0) + delta.seconds

        numticks = 5

        # self._freq = YEARLY
        interval = 1
        bymonth = 1
        bymonthday = 1
        byhour = 0
        byminute = 0
        bysecond = 0

        if numYears >= numticks:
            self._freq = YEARLY
        elif numMonths >= numticks:
            self._freq = MONTHLY
            bymonth = list(range(1, 13))
            if (0 <= numMonths) and (numMonths <= 14):
                interval = 1  # show every month
            elif (15 <= numMonths) and (numMonths <= 29):
                interval = 3  # show every 3 months
            elif (30 <= numMonths) and (numMonths <= 44):
                interval = 4  # show every 4 months
            else:  # 45 <= numMonths <= 59
                interval = 6  # show every 6 months
        elif numDays >= numticks:
            self._freq = DAILY
            bymonth = None
            bymonthday = list(range(1, 32))
            if (0 <= numDays) and (numDays <= 9):
                interval = 1  # show every day
            elif (10 <= numDays) and (numDays <= 19):
                interval = 2  # show every 2 days
            elif (20 <= numDays) and (numDays <= 35):
                interval = 3  # show every 3 days
            elif (36 <= numDays) and (numDays <= 80):
                interval = 7  # show every 1 week
            else:  # 100 <= numDays <= ~150
                interval = 14  # show every 2 weeks
        elif numHours >= numticks:
            self._freq = HOURLY
            bymonth = None
            bymonthday = None
            byhour = list(range(0, 24))  # show every hour
            if (0 <= numHours) and (numHours <= 14):
                interval = 1  # show every hour
            elif (15 <= numHours) and (numHours <= 30):
                interval = 2  # show every 2 hours
            elif (30 <= numHours) and (numHours <= 45):
                interval = 3  # show every 3 hours
            elif (45 <= numHours) and (numHours <= 68):
                interval = 4  # show every 4 hours
            elif (68 <= numHours) and (numHours <= 90):
                interval = 6  # show every 6 hours
            else:  # 90 <= numHours <= 120
                interval = 12  # show every 12 hours
        elif numMinutes >= numticks:
            self._freq = MINUTELY
            bymonth = None
            bymonthday = None
            byhour = None
            byminute = list(range(0, 60))
            if numMinutes > (10.0 * numticks):
                interval = 10
            # end if
        elif numSeconds >= numticks:
            self._freq = SECONDLY
            bymonth = None
            bymonthday = None
            byhour = None
            byminute = None
            bysecond = list(range(0, 60))
            if numSeconds > (10.0 * numticks):
                interval = 10
            # end if
        else:
            # do what?
            #   microseconds as floats, but floats from what reference point?
            pass

        rrule = rrulewrapper(
            self._freq,
            interval=interval,
            dtstart=dmin,
            until=dmax,
            bymonth=bymonth,
            bymonthday=bymonthday,
            byhour=byhour,
            byminute=byminute,
            bysecond=bysecond,
        )

        locator = RRuleLocator(rrule, self.tz)
        locator.set_axis(self.axis)

        locator.set_view_interval(*self.axis.get_view_interval())
        locator.set_data_interval(*self.axis.get_data_interval())
        return locator


def pretty_float(num):

    if num > 1000:
        return comma_format(int(num))

    try:
        floats = int(max(2 - max(numpy.floor(numpy.log(abs(num) + 1e-3) / numpy.log(10.0)), 0), 0))
    except Exception:
        floats = 2
    format = "%." + str(floats) + "f"
    if isinstance(num, tuple):
        return format % float(num[0])
    else:
        try:
            retval = format % float(num)
        except Exception:
            raise Exception("Unable to convert %s into a float." % (str(num)))
        return retval


def statistics(results, span=None, is_timestamp=False):
    results = dict(results)
    if span is not None:
        parsed_data = {}
        min_key = min(results)
        max_key = max(results)
        for i in range(min_key, max_key + span, span):
            if i in results:
                parsed_data[i] = results[i]
                del results[i]
            else:
                parsed_data[i] = 0.0
        if len(results) > 0:
            raise Exception("Unable to use all the values for the statistics")
    else:
        parsed_data = results
    values = list(parsed_data.values())
    data_min = min(values)
    data_max = max(values)
    data_avg = numpy.average(values)
    if is_timestamp:
        current_time = max(parsed_data)
        data_current = parsed_data[current_time]
        return data_min, data_max, data_avg, data_current
    else:
        return data_min, data_max, data_avg


def makeDataFromCSV(csv):
    """Generate plot data dictionary from a csv file or string"""

    if os.path.exists(csv):
        with open(csv, "r") as fdata:
            flines = fdata.readlines()
    else:
        flines = csv.split("\n")

    graph_data = {}
    labels = flines[0].strip().split(",")
    if len(labels) == 2:
        # simple plot data
        for line in flines:
            line = line.strip()
            if line[0] != "#":
                key, value = line.split(",")
                graph_data[key] = value

    elif len(flines) == 2:
        values = flines[1].strip().split(",")
        for key, value in zip(labels, values):
            graph_data[key] = value

    elif len(labels) > 2:
        # stacked graph data
        del labels[0]
        del flines[0]
        for label in labels:
            plot_data = {}
            index = labels.index(label) + 1
            for line in flines:
                values = line.strip().split(",")
                value = values[index].strip()
                # if value:
                plot_data[values[0]] = values[index]
                # else:
                # plot_data[values[0]] = '0.'
                # pass
            graph_data[label] = dict(plot_data)

    return graph_data


def darkenColor(color, factor=2):

    c1 = int(color[1:3], 16)
    c2 = int(color[3:5], 16)
    c3 = int(color[5:7], 16)

    c1 /= factor
    c2 /= factor
    c3 /= factor

    result = "#" + (
        str(hex(c1)).replace("0x", "").zfill(2)
        + str(hex(c2)).replace("0x", "").zfill(2)
        + str(hex(c3)).replace("0x", "").zfill(2)
    )
    return result



#!/usr/bin/env python
########################################################################
# File :    dirac-jobexec
# Author :  Stuart Paterson
########################################################################
""" The dirac-jobexec script is equipped to execute workflows that
    are specified via their XML description.  The main client of
    this script is the Job Wrapper.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import os
import os.path
import sys

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Register workflow parameter switch
    Script.registerSwitch("p:", "parameter=", "Parameters that are passed directly to the workflow")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("jobXMLfile: specify path to the Job XML file description")
    Script.parseCommandLine()

    # from DIRAC.Core.Workflow.Parameter import *
    from DIRAC import gLogger
    from DIRAC.Core.Workflow.Workflow import fromXMLFile
    from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
    from DIRAC.AccountingSystem.Client.DataStoreClient import DataStoreClient
    from DIRAC.RequestManagementSystem.Client.Request import Request

    # Forcing the current directory to be the first in the PYTHONPATH
    sys.path.insert(0, os.path.realpath("."))
    gLogger.showHeaders(True)

    def jobexec(jobxml, wfParameters):
        jobfile = os.path.abspath(jobxml)
        if not os.path.exists(jobfile):
            gLogger.warn("Path to specified workflow %s does not exist" % (jobfile))
            sys.exit(1)
        workflow = fromXMLFile(jobfile)
        gLogger.debug(workflow)
        code = workflow.createCode()
        gLogger.debug(code)
        jobID = 0
        if "JOBID" in os.environ:
            jobID = os.environ["JOBID"]
            gLogger.info("DIRAC JobID %s is running at site %s" % (jobID, DIRAC.siteName()))

        workflow.addTool("JobReport", JobReport(jobID))
        workflow.addTool("AccountingReport", DataStoreClient())
        workflow.addTool("Request", Request())

        # Propagate the command line parameters to the workflow if any
        for pName, pValue in wfParameters.items():
            workflow.setValue(pName, pValue)

        # Propagate the command line parameters to the workflow module instances of each step
        for stepdefinition in workflow.step_definitions.values():
            for moduleInstance in stepdefinition.module_instances:
                for pName, pValue in wfParameters.items():
                    if moduleInstance.parameters.find(pName):
                        moduleInstance.parameters.setValue(pName, pValue)

        return workflow.execute()

    positionalArgs = Script.getPositionalArgs()
    if len(positionalArgs) != 1:
        gLogger.debug("Positional arguments were %s" % (positionalArgs))
        DIRAC.abort(1, "Must specify the Job XML file description")

    if "JOBID" in os.environ:
        gLogger.info("JobID: %s" % (os.environ["JOBID"]))

    jobXMLfile = positionalArgs[0]
    parList = Script.getUnprocessedSwitches()
    parDict = {}
    for switch, parameter in parList:
        if switch == "p":
            name, value = parameter.split("=")
            value = value.strip()

            # The comma separated list in curly brackets is interpreted as a list
            if value.startswith("{"):
                value = value[1:-1].replace('"', "").replace(" ", "").split(",")
                value = ";".join(value)

            parDict[name] = value

    gLogger.debug("PYTHONPATH:\n%s" % ("\n".join(sys.path)))
    jobExec = jobexec(jobXMLfile, parDict)
    if not jobExec["OK"]:
        gLogger.debug("Workflow execution finished with errors, exiting")
        if jobExec["Errno"]:
            sys.exit(jobExec["Errno"])
        else:
            sys.exit(1)
    else:
        gLogger.debug("Workflow execution successful, exiting")
        sys.exit(0)


if __name__ == "__main__":
    main()



"""  The CountryMapping module performs the necessary CS gymnastics to resolve country codes  """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gConfig, S_OK, S_ERROR


def getCountryMapping(country):
    """Determines the associated country from the country code"""
    mappedCountries = [country]
    while True:
        mappedCountry = gConfig.getValue("/Resources/Countries/%s/AssignedTo" % country, country)
        if mappedCountry == country:
            break
        elif mappedCountry in mappedCountries:
            return S_ERROR("Circular mapping detected for %s" % country)
        else:
            country = mappedCountry
            mappedCountries.append(mappedCountry)
    return S_OK(mappedCountry)


def getCountryMappingTier1(country):
    """Returns the Tier1 site mapped to a country code"""
    res = getCountryMapping(country)
    if not res["OK"]:
        return res
    mappedCountry = res["Value"]
    tier1 = gConfig.getValue("/Resources/Countries/%s/Tier1" % mappedCountry, "")
    if not tier1:
        return S_ERROR("No Tier1 assigned to %s" % mappedCountry)
    return S_OK(tier1)



""" PDP

  PDP ( PolicyDecisionPoint ) is the back-end for the PolicySystem. It discovers
  the policies, finds the best match, evaluates them, merges their results taking
  the most penalizing one, computes the set of actions to be triggered and returns
  all the information to the PEP which will enforce the actions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyCaller import PolicyCaller
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import RSSMachine
from DIRAC.ResourceStatusSystem.Utilities import RssConfiguration
from DIRAC.ResourceStatusSystem.Utilities.InfoGetter import getPolicyActionsThatApply, getPoliciesThatApply

__RCSID__ = "$Id$"


class PDP(object):
    """PDP ( Policy Decision Point )"""

    def __init__(self, clients=None):
        """Constructor.

        examples:
          >>> pdp  = PDP( None )
          >>> pdp1 = PDP( {} )
          >>> pdp2 = PDP( { 'Client1' : Client1Object } )

        :Parameters:
          **clients** - [ None, `dict` ]
            dictionary with Clients to be used in the Commands. If None, the Commands
            will create their own clients.

        """

        # decision parameters used to match policies and actions
        self.decisionParams = None

        # Helpers to discover policies and RSS metadata in CS
        self.pCaller = PolicyCaller(clients)

        # RSS State Machine, used to calculate most penalizing state while merging them
        self.rssMachine = RSSMachine("Unknown")

        self.log = gLogger.getSubLogger("PDP")

    def setup(self, decisionParams=None):
        """method that sanitizes the decisionParams and ensures that at least it has
        the keys in `standardParamsDict`. This will be relevant while doing the matching
        with the RSS Policies configuration in the CS.
        There is one key-value pair, `active` which is added on this method. This allows
        policies to be de-activated from the CS, changing their active matchParam to
        something else than `Active`.

        examples:
          >>> pdp.setup( None )
          >>> self.decisionParams
              { 'element' : None, 'name' : None, ... }
          >>> pdp.setup( { 'element' : 'AnElement' } )
          >>> self.decisionParams
              { 'element' : 'AnElement', 'name' : None, ... }
          >>> pdp.setup( { 'NonStandardKey' : 'Something' } )
          >>> self.decisionParams
              { 'NonStandardKey' : 'Something', 'element' : None,... }

        :Parameters:
          **decisionParams** - [ None, `dict` ]
            dictionary with the parameters to be matched with the RSS Policies configuration
            in the CS.

        """

        standardParamsDict = {
            "element": None,
            "name": None,
            "elementType": None,
            "statusType": None,
            "status": None,
            "reason": None,
            "tokenOwner": None,
            # Last parameter allows policies to be de-activated
            "active": "Active",
        }

        if decisionParams is not None:
            standardParamsDict.update(decisionParams)
            if standardParamsDict["element"] is not None:
                self.log = gLogger.getSubLogger("PDP/%s" % standardParamsDict["element"])
                if standardParamsDict["name"] is not None:
                    self.log = gLogger.getSubLogger(
                        "PDP/%s/%s" % (standardParamsDict["element"], standardParamsDict["name"])
                    )
                    self.log.verbose(
                        "Setup - statusType: %s, status: %s"
                        % (standardParamsDict["statusType"], standardParamsDict["status"])
                    )
            self.decisionParams = standardParamsDict

    def takeDecision(self):
        """main PDP method which does all the work. If firstly finds all the policies
        defined in the CS that match <self.decisionParams> and runs them. Once it has
        all the singlePolicyResults, it combines them. Next step is action discovery:
        using a similar approach to the one used to discover the policies, but also
        taking into account the single policy results and their combined result, finds
        the actions to be triggered and returns.

        examples:
          >>> list(pdp.takeDecision()['Value'])
              ['singlePolicyResults', 'policyCombinedResult', 'decisionParams']
          >>> pdp.takeDecision()['Value']['singlePolicyResults']
              [ { 'Status' : 'Active',
                  'Reason' : 'blah',
                  'Policy' : { 'name'        : 'AlwaysActiveForResource',
                               'type'        : 'AlwaysActive',
                               'module'      : 'AlwaysActivePolicy',
                               'description' : 'This is the AlwaysActive policy'
                               'command'     : None,
                               'args'        : {}
                             }
                }, ... ]
          >>> pdp.takeDecision()['Value']['policyCombinedResult']
              { 'Status'       : 'Active',
                'Reason'       : 'blah ###',
                'PolicyAction' : [ ( 'policyActionName1', 'policyActionType1' ), ... ]
              }

        :return: S_OK( { 'singlePolicyResults'  : `list`,
                         'policyCombinedResult' : `dict`,
                         'decisionParams'      : `dict` } ) / S_ERROR

        """
        if self.decisionParams is None:
            return S_OK({"singlePolicyResults": [], "policyCombinedResult": {}, "decisionParams": self.decisionParams})

        self.log.verbose("Taking decision")

        # Policies..................................................................

        # Get policies that match self.decisionParams
        policiesThatApply = getPoliciesThatApply(self.decisionParams)
        if not policiesThatApply["OK"]:
            return policiesThatApply
        policiesThatApply = policiesThatApply["Value"]
        self.log.verbose("Policies that apply: %s" % ", ".join([po["name"] for po in policiesThatApply]))

        # Evaluate policies
        singlePolicyResults = self._runPolicies(policiesThatApply)
        if not singlePolicyResults["OK"]:
            return singlePolicyResults
        singlePolicyResults = singlePolicyResults["Value"]
        self.log.verbose("Single policy results: %s" % singlePolicyResults)

        # Combine policies and get most penalizing status ( see RSSMachine )
        policyCombinedResults = self._combineSinglePolicyResults(singlePolicyResults)
        if not policyCombinedResults["OK"]:
            return policyCombinedResults
        policyCombinedResults = policyCombinedResults["Value"]
        self.log.verbose("Combined policy result: %s" % policyCombinedResults)

        # Actions...................................................................

        policyActionsThatApply = getPolicyActionsThatApply(
            self.decisionParams, singlePolicyResults, policyCombinedResults
        )
        if not policyActionsThatApply["OK"]:
            return policyActionsThatApply
        policyActionsThatApply = policyActionsThatApply["Value"]
        self.log.verbose("Policy actions that apply: %s" % ",".join(pata[0] for pata in policyActionsThatApply))

        policyCombinedResults["PolicyAction"] = policyActionsThatApply

        return S_OK(
            {
                "singlePolicyResults": singlePolicyResults,
                "policyCombinedResult": policyCombinedResults,
                "decisionParams": self.decisionParams,
            }
        )

    def _runPolicies(self, policies):
        """Given a list of policy dictionaries, loads them making use of the PolicyCaller
        and evaluates them. This method requires to have run setup previously.

        examples:
          >>> pdp._runPolicies([])['Value']
              []
          >>> policyDict = { 'name'        : 'AlwaysActiveResource',
                             'type'        : 'AlwaysActive',
                             'args'        : None,
                             'description' : 'This is the AlwaysActive policy',
                             'module'      : 'AlwaysActivePolicy',
                             'command'     : None }
          >>> pdp._runPolicies([ policyDict, ... ] )['Value']
              [ { 'Status' : 'Active', 'Reason' : 'blah', 'Policy' : policyDict }, ... ]

        :Parameters:
          **policies** - `list( dict )`
            list of dictionaries containing the policies selected to be run. Check the
            examples to get an idea of how the policy dictionaries look like.

        :return: S_OK() / S_ERROR

        """

        policyInvocationResults = []

        # Gets all valid status for RSS to avoid misconfigured policies returning statuses
        # that RSS does not understand.
        validStatus = self.rssMachine.getStates()

        for policyDict in policies:

            # Load and evaluate policy described in <policyDict> for element described
            # in <self.decisionParams>
            policyInvocationResult = self.pCaller.policyInvocation(self.decisionParams, policyDict)
            if not policyInvocationResult["OK"]:
                # We should never enter this line ! Just in case there are policies
                # missconfigured !
                _msg = "runPolicies no OK: %s" % policyInvocationResult
                self.log.error(_msg)
                return S_ERROR(_msg)

            policyInvocationResult = policyInvocationResult["Value"]

            # Sanity Checks ( they should never happen ! )
            if "Status" not in policyInvocationResult:
                _msg = "runPolicies (no Status): %s" % policyInvocationResult
                self.log.error(_msg)
                return S_ERROR(_msg)

            if not policyInvocationResult["Status"] in validStatus:
                _msg = "runPolicies ( not valid status ) %s" % policyInvocationResult["Status"]
                self.log.error(_msg)
                return S_ERROR(_msg)

            if "Reason" not in policyInvocationResult:
                _msg = "runPolicies (no Reason): %s" % policyInvocationResult
                self.log.error(_msg)
                return S_ERROR(_msg)

            policyInvocationResults.append(policyInvocationResult)

        return S_OK(policyInvocationResults)

    def _combineSinglePolicyResults(self, singlePolicyRes):
        """method that merges all the policies results into a combined one, which
        will be the most penalizing status and the reasons of the single policy
        results that returned the same penalizing status. All the rest, are ignored.
        If there are no single policy results, it is returned `Unknown` state. While
        combining policies, the ones containing the option `doNotCombine` are ignored.

        examples:
          >>> pdp._combineSingePolicyResults( [] )['Value']
              { 'Status' : 'Unknown', 'Reason' : 'No policy ..' }
          >>> pdp._combineSingePolicyResults( [ { 'Status' : 'Active', 'Reason' : 'blah', 'Policy' : policyDict } ] )
              { 'Status' : 'Active', 'Reason' : 'blah' }
          >>> pdp._combineSingePolicyResults( [ { 'Status' : 'Active', 'Reason' : 'blah', 'Policy' : policyDict },
                                                { 'Status' : 'Banned', 'Reason' : 'blah 2', 'Policy' : policyDict2 } ] )
              { 'Status' : 'Banned', 'Reason' : 'blah 2' }
          >>> pdp._combineSingePolicyResults( [ { 'Status' : 'Active', 'Reason' : 'blah', 'Policy' : policyDict },
                                                { 'Status' : 'Active', 'Reason' : 'blah 2', 'Policy' : policyDict2 } ] )
              { 'Status' : 'Banned', 'Reason' : 'blah ### blah 2' }

        :Parameters:
          **singlePolicyRes** - `list( dict )`
            list with every single policy result to be combined ( see _runPolicy for more details )

        :return: S_OK( dict( Status, Reason ) | S_ERROR

        """

        # Dictionary to be returned
        policyCombined = {
            "Status": "Unknown",  # default, it should be overridden by the policies, if they exist
            "Reason": "",
        }

        # If there are no policyResults, we return Unknown
        if not singlePolicyRes:
            policyCombined["Reason"] = (
                "No policy applies to %(element)s, %(name)s, %(elementType)s" % self.decisionParams
            )
            self.log.warn(policyCombined["Reason"])
            return S_OK(policyCombined)

        # We set the rssMachine on the current state ( ensures it is a valid one )
        # FIXME: probably this check can be done at takeDecision
        machineStatus = self.rssMachine.setState(self.decisionParams["status"], noWarn=True)
        if not machineStatus["OK"]:
            return machineStatus

        # Discard all single policy results which belongs to policies that have set
        # the option `doNotCombine` in the CS
        policiesToCombine = self._findPoliciesToCombine(singlePolicyRes)

        # Sort policy results using ther statuses by most restrictive ( lower level first )
        self.rssMachine.orderPolicyResults(policiesToCombine)

        # As they have been sorted by most restrictive status, the first one is going
        # to be our candidate new state. Let's ask the RSSMachine if it allows us to
        # make such transition.
        candidateState = policiesToCombine[0]["Status"]
        nextState = self.rssMachine.getNextState(candidateState)

        if not nextState["OK"]:
            return nextState
        nextState = nextState["Value"]
        # most restrictive status defines the VO affected. VO='all' will affect all VOs
        policyCombined["VO"] = policiesToCombine[0].get("VO", "all")

        # If the RssMachine does not accept the candidate, return forcing message
        if candidateState != nextState:

            policyCombined["Status"] = nextState
            policyCombined["Reason"] = "RssMachine forced status %s to %s" % (candidateState, nextState)
            return S_OK(policyCombined)

        # If the RssMachine accepts the candidate, just concatenate the reasons
        for policyRes in policiesToCombine:

            if policyRes["Status"] == nextState:
                policyCombined["Reason"] += "%s ###" % policyRes["Reason"]

        policyCombined["Status"] = nextState

        return S_OK(policyCombined)

    def _findPoliciesToCombine(self, singlePolicyRes):
        """method that iterates over the single policy results and checks the CS
        configuration of the policies looking for the option 'doNotCombine'. If it is
        present, that single policy result is discarded.

        :Parameters:
          **singlePolicyRes** - `list( dict )`
            list with every single policy result to be combined ( see _runPolicy for more details )

        :return: `list( dict )`

        """

        # Get policies configuration from the CS. We want to exclude the policies that
        # have set the option `doNotCombine` from this process.
        policiesConfiguration = RssConfiguration.getPolicies()
        if not policiesConfiguration["OK"]:
            return policiesConfiguration
        policiesConfiguration = policiesConfiguration["Value"]

        # Function that let's us know if we should combine the result of a single policy
        # or not.
        def combinePolicy(policyResult):
            # Extract policy name from the dictionary returned by PolicyCaller
            policyName = policyResult["Policy"]["name"]
            try:
                # If doNotCombineResult is defined, the policy is not taken into account
                # to create the combined result. However, the single policy result remains
                _ = policiesConfiguration[policyName]["doNotCombineResult"]
                return False
            except KeyError:
                return True

        # Make a list of policies of which we want to merge their results
        return [policyResult for policyResult in singlePolicyRes if combinePolicy(policyResult)]


# ...............................................................................
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" SandboxHandler is the implementation of the Sandbox service
    in the DISET framework
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time
import threading
import tempfile

import six
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Security import Locations, Properties, X509Certificate
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Service.StorageElementHandler import getDiskSpace
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.Resources.Storage.StorageElement import StorageElement


class SandboxStoreHandler(RequestHandler):

    __purgeCount = -1
    __purgeLock = threading.Lock()
    __purgeWorking = False

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB object"""
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.SandboxMetadataDB", "SandboxMetadataDB")
            if not result["OK"]:
                return result
            cls.sandboxDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)
        return S_OK()

    def initialize(self):
        self.__backend = self.getCSOption("Backend", "local")
        self.__localSEName = self.getCSOption("LocalSE", "SandboxSE")
        self.__maxUploadBytes = self.getCSOption("MaxSandboxSizeMiB", 10) * 1048576
        if self.__backend.lower() == "local" or self.__backend == self.__localSEName:
            self.__useLocalStorage = True
            self.__seNameToUse = self.__localSEName
        else:
            self.__useLocalStorage = False
            self.__externalSEName = self.__backend
            self.__seNameToUse = self.__backend
        # Execute the purge once every 1000 calls
        SandboxStoreHandler.__purgeCount += 1
        if SandboxStoreHandler.__purgeCount > self.getCSOption("QueriesBeforePurge", 1000):
            SandboxStoreHandler.__purgeCount = 0
        if SandboxStoreHandler.__purgeCount == 0:
            threading.Thread(target=self.purgeUnusedSandboxes).start()

    def __getSandboxPath(self, md5):
        """Generate the sandbox path"""
        # prefix = self.getCSOption( "SandboxPrefix", "SandBox" )
        prefix = "SandBox"
        credDict = self.getRemoteCredentials()
        if Properties.JOB_SHARING in credDict["properties"]:
            idField = credDict["group"]
        else:
            idField = "%s.%s" % (credDict["username"], credDict["group"])
        pathItems = ["/", prefix, idField[0], idField]
        pathItems.extend([md5[0:3], md5[3:6], md5])
        return os.path.join(*pathItems)

    def transfer_fromClient(self, fileId, token, fileSize, fileHelper):
        """
        Receive a file as a sandbox
        """

        if self.__maxUploadBytes and fileSize > self.__maxUploadBytes:
            fileHelper.markAsTransferred()
            return S_ERROR("Sandbox is too big. Please upload it to a grid storage element")

        if isinstance(fileId, (list, tuple)):
            if len(fileId) > 1:
                assignTo = fileId[1]
                fileId = fileId[0]
            else:
                return S_ERROR("File identified tuple has to have length greater than 1")
        else:
            assignTo = {}

        extPos = fileId.find(".tar")
        if extPos > -1:
            extension = fileId[extPos + 1 :]
            aHash = fileId[:extPos]
        else:
            extension = ""
            aHash = fileId
        gLogger.info("Upload requested for %s [%s]" % (aHash, extension))

        credDict = self.getRemoteCredentials()
        sbPath = self.__getSandboxPath("%s.%s" % (aHash, extension))
        # Generate the location
        result = self.__generateLocation(sbPath)
        if not result["OK"]:
            return result
        seName, sePFN = result["Value"]

        result = self.sandboxDB.getSandboxId(seName, sePFN, credDict["username"], credDict["group"])
        if result["OK"]:
            gLogger.info("Sandbox already exists. Skipping upload")
            fileHelper.markAsTransferred()
            sbURL = "SB:%s|%s" % (seName, sePFN)
            assignTo = dict([(key, [(sbURL, assignTo[key])]) for key in assignTo])
            result = self.export_assignSandboxesToEntities(assignTo)
            if not result["OK"]:
                return result
            return S_OK(sbURL)

        if self.__useLocalStorage:
            hdPath = self.__sbToHDPath(sbPath)
        else:
            hdPath = False
        # Write to local file
        result = self.__networkToFile(fileHelper, hdPath)
        if not result["OK"]:
            gLogger.error("Error while receiving sandbox file", "%s" % result["Message"])
            return result
        hdPath = result["Value"]
        gLogger.info("Wrote sandbox to file %s" % hdPath)
        # Check hash!
        if fileHelper.getHash() != aHash:
            self.__secureUnlinkFile(hdPath)
            gLogger.error("Hashes don't match! Client defined hash is different with received data hash!")
            return S_ERROR("Hashes don't match!")
        # If using remote storage, copy there!
        if not self.__useLocalStorage:
            gLogger.info("Uploading sandbox to external storage")
            result = self.__copyToExternalSE(hdPath, sbPath)
            self.__secureUnlinkFile(hdPath)
            if not result["OK"]:
                return result
            sbPath = result["Value"][1]
        # Register!
        gLogger.info("Registering sandbox in the DB with", "SB:%s|%s" % (self.__seNameToUse, sbPath))
        result = self.sandboxDB.registerAndGetSandbox(
            credDict["username"],
            credDict["DN"],
            credDict["group"],
            self.__seNameToUse,
            sbPath,
            fileHelper.getTransferedBytes(),
        )
        if not result["OK"]:
            self.__secureUnlinkFile(hdPath)
            return result

        sbURL = "SB:%s|%s" % (self.__seNameToUse, sbPath)
        assignTo = dict([(key, [(sbURL, assignTo[key])]) for key in assignTo])
        result = self.export_assignSandboxesToEntities(assignTo)
        if not result["OK"]:
            return result
        return S_OK(sbURL)

    def transfer_bulkFromClient(self, fileId, token, _fileSize, fileHelper):
        """Receive files packed into a tar archive by the fileHelper logic.
        token is used for access rights confirmation.
        """
        result = self.__networkToFile(fileHelper)
        if not result["OK"]:
            return result
        tmpFilePath = result["OK"]
        gLogger.info("Got Sandbox to local storage", tmpFilePath)

        extension = fileId[fileId.find(".tar") + 1 :]
        sbPath = "%s.%s" % (self.__getSandboxPath(fileHelper.getHash()), extension)
        gLogger.info("Sandbox path will be", sbPath)
        # Generate the location
        result = self.__generateLocation(sbPath)
        if not result["OK"]:
            return result
        seName, sePFN = result["Value"]
        # Register in DB
        credDict = self.getRemoteCredentials()
        result = self.sandboxDB.getSandboxId(seName, sePFN, credDict["username"], credDict["group"])
        if result["OK"]:
            return S_OK("SB:%s|%s" % (seName, sePFN))

        result = self.sandboxDB.registerAndGetSandbox(
            credDict["username"], credDict["DN"], credDict["group"], seName, sePFN, fileHelper.getTransferedBytes()
        )
        if not result["OK"]:
            self.__secureUnlinkFile(tmpFilePath)
            return result
        sbid, _newSandbox = result["Value"]
        gLogger.info("Registered in DB", "with SBId %s" % sbid)

        result = self.__moveToFinalLocation(tmpFilePath, sbPath)
        self.__secureUnlinkFile(tmpFilePath)
        if not result["OK"]:
            gLogger.error("Could not move sandbox to final destination", result["Message"])
            return result

        gLogger.info("Moved to final destination")

        # Unlink temporal file if it's there
        self.__secureUnlinkFile(tmpFilePath)
        return S_OK("SB:%s|%s" % (seName, sePFN))

    def __generateLocation(self, sbPath):
        """
        Generate the location string
        """
        if self.__useLocalStorage:
            return S_OK((self.__localSEName, sbPath))
        # It's external storage
        storageElement = StorageElement(self.__externalSEName)
        res = storageElement.isValid()
        if not res["OK"]:
            errStr = "Failed to instantiate destination StorageElement"
            gLogger.error(errStr, self.__externalSEName)
            return S_ERROR(errStr)
        result = storageElement.getURL(sbPath)
        if not result["OK"] or sbPath not in result["Value"]["Successful"]:
            errStr = "Failed to generate PFN"
            gLogger.error(errStr, self.__externalSEName)
            return S_ERROR(errStr)
        destPfn = result["Value"]["Successful"][sbPath]
        return S_OK((self.__externalSEName, destPfn))

    def __sbToHDPath(self, sbPath):
        while sbPath and sbPath[0] == "/":
            sbPath = sbPath[1:]
        basePath = self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes")
        return os.path.join(basePath, sbPath)

    def __networkToFile(self, fileHelper, destFileName=False):
        """
        Dump incoming network data to temporal file
        """
        tfd = None
        if not destFileName:
            try:
                tfd, destFileName = tempfile.mkstemp(prefix="DSB.")
                tfd.close()
            except Exception as e:
                gLogger.error("%s" % repr(e).replace(",)", ")"))
                return S_ERROR("Cannot create temporary file")

        destFileName = os.path.realpath(destFileName)
        mkDir(os.path.dirname(destFileName))

        try:
            if tfd is not None:
                fd = tfd
            else:
                fd = open(destFileName, "wb")
            result = fileHelper.networkToDataSink(fd, maxFileSize=self.__maxUploadBytes)
            fd.close()
        except Exception as e:
            gLogger.error(
                "Cannot open to write destination file", "%s: %s" % (destFileName, repr(e).replace(",)", ")"))
            )
            return S_ERROR("Cannot open to write destination file")
        if not result["OK"]:
            return result
        return S_OK(destFileName)

    def __secureUnlinkFile(self, filePath):
        try:
            os.unlink(filePath)
        except Exception as e:
            gLogger.warn("Could not unlink file %s: %s" % (filePath, repr(e).replace(",)", ")")))
            return False
        return True

    def __moveToFinalLocation(self, localFilePath, sbPath):
        if self.__useLocalStorage:
            hdFilePath = self.__sbToHDPath(sbPath)
            result = S_OK((self.__localSEName, sbPath))
            if os.path.isfile(hdFilePath):
                gLogger.info("There was already a sandbox with that name, skipping copy", sbPath)
            else:
                hdDirPath = os.path.dirname(hdFilePath)
                mkDir(hdDirPath)
                try:
                    os.rename(localFilePath, hdFilePath)
                except OSError as e:
                    errMsg = "Cannot move temporal file to final path"
                    gLogger.error(errMsg, repr(e).replace(",)", ")"))
                    result = S_ERROR(errMsg)
        else:
            result = self.__copyToExternalSE(localFilePath, sbPath)

        return result

    def __copyToExternalSE(self, localFilePath, sbPath):
        """
        Copy uploaded file to external SE
        """
        try:
            dm = DataManager()
            result = dm.put(sbPath, localFilePath, self.__externalSEName)
            if not result["OK"]:
                return result
            if "Successful" not in result["Value"]:
                gLogger.verbose("Oops, no successful transfers there", str(result))
                return S_ERROR("RM returned OK to the action but no successful transfers were there")
            okTrans = result["Value"]["Successful"]
            if sbPath not in okTrans:
                gLogger.verbose("Ooops, SB transfer wasn't in the successful ones", str(result))
                return S_ERROR("RM returned OK to the action but SB transfer wasn't in the successful ones")
            return S_OK((self.__externalSEName, okTrans[sbPath]))
        except Exception as e:
            gLogger.error("Error while moving sandbox to SE", "%s" % repr(e).replace(",)", ")"))
            return S_ERROR("Error while moving sandbox to SE")

    ##################
    # Assigning sbs to jobs

    types_assignSandboxesToEntities = [dict]

    def export_assignSandboxesToEntities(self, enDict, ownerName="", ownerGroup="", entitySetup=False):
        """
        Assign sandboxes to jobs.
        Expects a dict of { entityId : [ ( SB, SBType ), ... ] }
        """
        if not entitySetup:
            entitySetup = self.serviceInfoDict["clientSetup"]
        credDict = self.getRemoteCredentials()
        return self.sandboxDB.assignSandboxesToEntities(
            enDict, credDict["username"], credDict["group"], entitySetup, ownerName, ownerGroup
        )

    ##################
    # Unassign sbs to jobs

    types_unassignEntities = [(list, tuple)]

    def export_unassignEntities(self, entitiesList, entitiesSetup=False):
        """
        Unassign a list of jobs
        """
        if not entitiesSetup:
            entitiesSetup = self.serviceInfoDict["clientSetup"]
        credDict = self.getRemoteCredentials()
        return self.sandboxDB.unassignEntities({entitiesSetup: entitiesList}, credDict["username"], credDict["group"])

    ##################
    # Getting assigned sandboxes

    types_getSandboxesAssignedToEntity = [six.string_types]

    def export_getSandboxesAssignedToEntity(self, entityId, entitySetup=False):
        """
        Get the sandboxes associated to a job and the association type
        """
        if not entitySetup:
            entitySetup = self.serviceInfoDict["clientSetup"]
        credDict = self.getRemoteCredentials()
        result = self.sandboxDB.getSandboxesAssignedToEntity(
            entityId, entitySetup, credDict["username"], credDict["group"]
        )
        if not result["OK"]:
            return result
        sbDict = {}
        for SEName, SEPFN, SBType in result["Value"]:  # pylint: disable=invalid-name
            if SBType not in sbDict:
                sbDict[SBType] = []
            sbDict[SBType].append("SB:%s|%s" % (SEName, SEPFN))
        return S_OK(sbDict)

    ##################
    # Disk space left management

    types_getFreeDiskSpace = []

    def export_getFreeDiskSpace(self):
        """Get the free disk space of the storage element
        If no size is specified, terabytes will be used by default.
        """

        return getDiskSpace(self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes"))

    types_getTotalDiskSpace = []

    def export_getTotalDiskSpace(self):
        """Get the total disk space of the storage element
        If no size is specified, terabytes will be used by default.
        """
        return getDiskSpace(self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes"), total=True)

    ##################
    # Download sandboxes

    def transfer_toClient(self, fileID, token, fileHelper):
        """Method to send files to clients.
        fileID is the local file name in the SE.
        token is used for access rights confirmation.
        """
        credDict = self.getRemoteCredentials()
        serviceURL = self.serviceInfoDict["URL"]
        filePath = fileID.replace(serviceURL, "")
        result = self.sandboxDB.getSandboxId(self.__localSEName, filePath, credDict["username"], credDict["group"])
        if not result["OK"]:
            return result
        sbId = result["Value"]
        self.sandboxDB.accessedSandboxById(sbId)
        # If it's a local file
        hdPath = self.__sbToHDPath(filePath)
        if not os.path.isfile(hdPath):
            return S_ERROR("Sandbox does not exist")
        result = fileHelper.getFileDescriptor(hdPath, "rb")
        if not result["OK"]:
            return S_ERROR("Failed to get file descriptor: %s" % result["Message"])
        fd = result["Value"]
        result = fileHelper.FDToNetwork(fd)
        fileHelper.oFile.close()
        return result

    ##################
    # Purge sandboxes

    def purgeUnusedSandboxes(self):
        # If a purge is already working skip
        SandboxStoreHandler.__purgeLock.acquire()
        try:
            if SandboxStoreHandler.__purgeWorking:
                if time.time() - SandboxStoreHandler.__purgeWorking < 86400:
                    gLogger.info("Sandbox purge still working")
                    return S_OK()
                gLogger.error("Sandbox purge took over 24 hours, it either died or needs optimising")
            SandboxStoreHandler.__purgeWorking = time.time()
        finally:
            SandboxStoreHandler.__purgeLock.release()

        gLogger.info("Purging sandboxes")
        result = self.sandboxDB.getUnusedSandboxes()
        if not result["OK"]:
            gLogger.error("Error while retrieving sandboxes to purge", result["Message"])
            SandboxStoreHandler.__purgeWorking = False
            return result
        sbList = result["Value"]
        gLogger.info("Got sandboxes to purge", "(%d)" % len(sbList))
        for i, (sbId, SEName, SEPFN) in enumerate(sbList):
            if i % 10000 == 0:
                gLogger.info("Purging", "%d out of %d" % (i, len(sbList)))
            self.__purgeSandbox(sbId, SEName, SEPFN)

        SandboxStoreHandler.__purgeWorking = False
        return S_OK()

    def __purgeSandbox(self, sbId, SEName, SEPFN):
        result = self.__deleteSandboxFromBackend(SEName, SEPFN)
        if not result["OK"]:
            gLogger.error("Cannot delete sandbox from backend", result["Message"])
            return
        result = self.sandboxDB.deleteSandboxes([sbId])
        if not result["OK"]:
            gLogger.error("Cannot delete sandbox from DB", result["Message"])

    def __deleteSandboxFromBackend(self, SEName, SEPFN):
        gLogger.info("Purging sandbox", "SB:%s|%s" % (SEName, SEPFN))
        if SEName != self.__localSEName:
            return self.__deleteSandboxFromExternalBackend(SEName, SEPFN)
        hdPath = self.__sbToHDPath(SEPFN)
        try:
            if not os.path.isfile(hdPath):
                return S_OK()
        except Exception as e:
            gLogger.error("Cannot perform isfile", "%s : %s" % (hdPath, repr(e).replace(",)", ")")))
            return S_ERROR("Error checking %s" % hdPath)
        try:
            os.unlink(hdPath)
        except Exception as e:
            gLogger.error("Cannot delete local sandbox", "%s : %s" % (hdPath, repr(e).replace(",)", ")")))
        while hdPath:
            hdPath = os.path.dirname(hdPath)
            gLogger.info("Checking if dir is empty", hdPath)
            try:
                if not os.path.isdir(hdPath):
                    break
                if os.listdir(hdPath):
                    break
                gLogger.info("Trying to clean dir", hdPath)
                # Empty dir!
                os.rmdir(hdPath)
            except Exception as e:
                gLogger.error("Cannot clean directory", "%s : %s" % (hdPath, repr(e).replace(",)", ")")))
                break
        return S_OK()

    def __deleteSandboxFromExternalBackend(self, SEName, SEPFN):
        if self.getCSOption("DelayedExternalDeletion", True):
            gLogger.info("Setting deletion request")
            try:

                # We need the hostDN used in order to pass these credentials to the
                # SandboxStoreDB..
                hostCertLocation, _ = Locations.getHostCertificateAndKeyLocation()
                hostCert = X509Certificate.X509Certificate()
                hostCert.loadFromFile(hostCertLocation)
                hostDN = hostCert.getSubjectDN().get("Value")

                # use the host authentication to fetch the data
                result = self.sandboxDB.getSandboxOwner(SEName, SEPFN, hostDN, "hosts")
                if not result["OK"]:
                    return result
                _owner, ownerDN, ownerGroup = result["Value"]

                request = Request()
                request.RequestName = "RemoteSBDeletion:%s|%s:%s" % (SEName, SEPFN, time.time())
                request.OwnerDN = ownerDN
                request.OwnerGroup = ownerGroup
                physicalRemoval = Operation()
                physicalRemoval.Type = "PhysicalRemoval"
                physicalRemoval.TargetSE = SEName
                fileToRemove = File()
                fileToRemove.PFN = SEPFN
                physicalRemoval.addFile(fileToRemove)
                request.addOperation(physicalRemoval)
                return ReqClient().putRequest(request)
            except Exception as e:
                gLogger.exception("Exception while setting deletion request")
                return S_ERROR("Cannot set deletion request: %s" % str(e))
        else:
            gLogger.info("Deleting external Sandbox")
            try:
                return StorageElement(SEName).removeFile(SEPFN)
            except Exception:
                gLogger.exception("RM raised an exception while trying to delete a remote sandbox")
                return S_ERROR("RM raised an exception while trying to delete a remote sandbox")



#!/bin/env python
"""
Tests for Bdii2CSAgent module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from mock import MagicMock as Mock, patch

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Agent import Bdii2CSAgent

MODNAME = "DIRAC.ConfigurationSystem.Agent.Bdii2CSAgent"

MAINBDII = {
    "site1": {"CEs": {"ce1": {"Queues": {"queue1": "SomeValues"}}}},
    "site2": {"CEs": {"ce2": {"Queues": {"queue2": "SomeValues"}}}},
}
ALTBDII = {
    "site2": {
        "CEs": {"ce2": {"Queues": {"queue2": "SomeOtherValues"}}, "ce2b": {"Queues": {"queue2b": "SomeOtherValues"}}}
    },
    "site3": {"CEs": {"ce3": {"Queues": {"queue3": "SomeValues"}}}},
}


class Bdii2CSTests(unittest.TestCase):
    def setUp(self):
        with patch("DIRAC.ConfigurationSystem.Agent.Bdii2CSAgent.AgentModule.__init__", new=Mock()), patch(
            "DIRAC.ConfigurationSystem.Agent.Bdii2CSAgent.AgentModule.am_getModuleParam",
            new=Mock(return_value="fullName"),
        ):
            self.agent = Bdii2CSAgent.Bdii2CSAgent(agentName="Configuration/testing", loadName="Configuration/testing")
        # as we ignore the init from the baseclass some agent variables might not be present so we set them here
        # in any case with this we can check that log is called with proper error messages
        self.agent.log = Mock()

    def tearDown(self):
        pass

    def test__getGlue2CEInfo_success(self):

        expectedResult = {}
        expectedResult.update(ALTBDII)
        expectedResult.update(MAINBDII)

        self.agent.alternativeBDIIs = ["server2"]
        with patch(
            MODNAME + ".getGlue2CEInfo",
            new=Mock(
                side_effect=[
                    S_OK(MAINBDII),
                    S_OK(ALTBDII),
                ]
            ),
        ) as infoMock:
            ret = self.agent._Bdii2CSAgent__getGlue2CEInfo("vo")  # pylint: disable=no-member
            infoMock.assert_any_call("vo", host=self.agent.host)
            infoMock.assert_any_call("vo", host="server2")
        self.assertTrue(ret["OK"])
        self.assertEqual(expectedResult, ret["Value"])
        self.assertEqual(ret["Value"]["site2"]["CEs"]["ce2"]["Queues"]["queue2"], "SomeValues")
        self.assertNotIn("ce2b", ret["Value"]["site2"]["CEs"])

    def test__getGlue2CEInfo_fail_10(self):

        self.agent.alternativeBDIIs = ["server2"]
        with patch(
            MODNAME + ".getGlue2CEInfo",
            new=Mock(
                side_effect=[
                    S_ERROR("error"),
                    S_OK(ALTBDII),
                ]
            ),
        ) as infoMock:
            ret = self.agent._Bdii2CSAgent__getGlue2CEInfo("vo")  # pylint: disable=no-member
            infoMock.assert_any_call("vo", host=self.agent.host)
            infoMock.assert_any_call("vo", host="server2")
            self.assertTrue(
                any(
                    "Failed getting information from default" in str(args)
                    for args in self.agent.log.error.call_args_list
                ),
                self.agent.log.error.call_args_list,
            )
        self.assertTrue(ret["OK"])
        self.assertEqual(ALTBDII, ret["Value"])

    def test__getGlue2CEInfo_fail_01(self):

        self.agent.alternativeBDIIs = ["server2"]
        with patch(
            MODNAME + ".getGlue2CEInfo",
            new=Mock(
                side_effect=[
                    S_OK(MAINBDII),
                    S_ERROR("error"),
                ]
            ),
        ) as infoMock:
            ret = self.agent._Bdii2CSAgent__getGlue2CEInfo("vo")  # pylint: disable=no-member
            infoMock.assert_any_call("vo", host="server2")
            infoMock.assert_any_call("vo", host="server2")
            self.assertTrue(
                any(
                    "Failed getting information from server2" in str(args)
                    for args in self.agent.log.error.call_args_list
                ),
                self.agent.log.error.call_args_list,
            )
        self.assertTrue(ret["OK"])
        self.assertEqual(MAINBDII, ret["Value"])

    def test__getGlue2CEInfo_fail_11(self):

        self.agent.alternativeBDIIs = ["server2"]
        with patch(
            MODNAME + ".getGlue2CEInfo",
            new=Mock(
                side_effect=[
                    S_ERROR("error1"),
                    S_ERROR("error2"),
                ]
            ),
        ) as infoMock:
            ret = self.agent._Bdii2CSAgent__getGlue2CEInfo("vo")  # pylint: disable=no-member
            infoMock.assert_any_call("vo", host=self.agent.host)
            infoMock.assert_any_call("vo", host="server2")
            self.assertTrue(
                any(
                    "Failed getting information from server2" in str(args)
                    for args in self.agent.log.error.call_args_list
                ),
                self.agent.log.error.call_args_list,
            )
            self.assertTrue(
                any(
                    "Failed getting information from default" in str(args)
                    for args in self.agent.log.error.call_args_list
                ),
                self.agent.log.error.call_args_list,
            )

        self.assertFalse(ret["OK"])
        self.assertIn("error1\nerror2", ret["Message"])


if __name__ == "__main__":
    SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(Bdii2CSTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



#!/usr/bin/env python
########################################################################
# File :   dirac-executor
# Author : Adria Casajus
########################################################################
"""
This is a script to launch DIRAC executors
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys

from DIRAC import gLogger
from DIRAC.Core.Base.ExecutorReactor import ExecutorReactor
from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["executor: specify which executor to run"])
    positionalArgs = Script.getPositionalArgs()
    localCfg = Script.localCfg

    if len(positionalArgs) == 1 and positionalArgs[0].find("/") > -1:
        mainName = positionalArgs[0]
    else:
        mainName = "Framework/MultiExecutor"

    localCfg.setConfigurationForExecutor(mainName)
    localCfg.addMandatoryEntry("/DIRAC/Setup")
    localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
    localCfg.addDefaultEntry("LogLevel", "INFO")
    localCfg.addDefaultEntry("LogColor", True)
    resultDict = localCfg.loadUserData()
    if not resultDict["OK"]:
        gLogger.fatal("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    includeExtensionErrors()
    executorReactor = ExecutorReactor()

    result = executorReactor.loadModules(positionalArgs)
    if not result["OK"]:
        gLogger.fatal("Error while loading executor", result["Message"])
        sys.exit(1)

    result = executorReactor.go()
    if not result["OK"]:
        gLogger.fatal(result["Message"])
        sys.exit(1)

    gLogger.notice("Graceful exit. Bye!")
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Check configuration options against the defaults in the ConfigTemplate.cfg files.

This script can help to discover discrepancies in the configuration:

  - Typos in option names
  - Removed options
  - Missing authorization settings

This script should be run by dirac administrators after major updates.

Usage:
  dirac-admin-check-config-options [options] -[MAUO] [-S <system>]
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
from pprint import pformat

from diraccfg import CFG
from DIRAC import gLogger, S_ERROR, S_OK, gConfig
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.List import fromChar

LOG = gLogger


class CheckConfig(object):
    """Compare the ConfigTemplate with current configuration."""

    def __init__(self):
        self.systems = None
        self.showModified = False
        self.showAdded = False
        self.showMissingSections = False
        self.showMissingOptions = False

    def _setSystems(self, val):
        self.systems = fromChar(val)
        return S_OK()

    def _setShowModified(self, _):
        self.showModified = True
        return S_OK()

    def _setShowAdded(self, _):
        self.showAdded = True
        return S_OK()

    def _setShowMissingSections(self, _):
        self.showMissingSections = True
        return S_OK()

    def _setShowMissingOptions(self, _):
        self.showMissingOptions = True
        return S_OK()

    def _setSwitches(self):
        Script.registerSwitch("S:", "system=", "Systems to check, by default all of them are checked", self._setSystems)
        Script.registerSwitch("M", "modified", "Show entries which differ from the default", self._setShowModified)
        Script.registerSwitch("A", "added", "Show entries which do not exist in ConfigTemplate", self._setShowAdded)
        Script.registerSwitch(
            "U",
            "missingSection",
            "Show sections which do not exist in the current configuration",
            self._setShowMissingSections,
        )
        Script.registerSwitch(
            "O",
            "missingOption",
            "Show options which do not exist in the current configuration",
            self._setShowMissingOptions,
        )

        Script.parseCommandLine(ignoreErrors=True)
        if not any([self.showModified, self.showAdded, self.showMissingSections, self.showMissingOptions]):
            LOG.error("\nERROR: Set at least one of the flags M A U O")
            Script.showHelp()

    def _check(self):
        """Obtain default configuration and current configuration and print the diff."""
        cfg = CFG()
        templateLocations = self._findConfigTemplates()
        for templatePath in templateLocations:
            cfgRes = self._parseConfigTemplate(templatePath, cfg)
            if cfgRes["OK"]:
                cfg = cfgRes["Value"]

        currentCfg = self._getCurrentConfig()
        if not currentCfg["OK"]:
            return
        currentCfg = currentCfg["Value"]
        diff = currentCfg.getModifications(cfg, ignoreOrder=True, ignoreComments=True)

        LOG.debug("*" * 80)
        LOG.debug("Default Configuration: %s" % str(cfg))
        LOG.debug("*" * 80)
        LOG.debug("Current Configuration: %s " % str(currentCfg))
        for entry in diff:
            self._printDiff(entry)

    def _parseConfigTemplate(self, templatePath, cfg=None):
        """Parse the ConfigTemplate.cfg files.

        :param str templatePath: path to the folder containing a ConfigTemplate.cfg file
        :param CFG cfg: cfg to merge with the systems config
        :returns: CFG object
        """
        cfg = CFG() if cfg is None else cfg

        system = os.path.split(templatePath.rstrip("/"))[1]
        if system.lower().endswith("system"):
            system = system[: -len("System")]

        if self.systems and system not in self.systems:
            return S_OK(cfg)

        templatePath = os.path.join(templatePath, "ConfigTemplate.cfg")
        if not os.path.exists(templatePath):
            return S_ERROR("File not found: %s" % templatePath)

        loadCfg = CFG()
        loadCfg.loadFromFile(templatePath)

        newCfg = CFG()
        newCfg.createNewSection("/%s" % system, contents=loadCfg)

        cfg = cfg.mergeWith(newCfg)

        return S_OK(cfg)

    @staticmethod
    def _findConfigTemplates():
        """Traverse folders in DIRAC and find ConfigTemplate.cfg files."""
        configTemplates = set()
        diracPath = os.environ.get("DIRAC")
        for baseDirectory, _subdirectories, files in os.walk(diracPath):
            if "ConfigTemplate.cfg" in files:
                configTemplates.add(baseDirectory)
        return configTemplates

    def _getCurrentConfig(self):
        """Return the current system configuration."""
        from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

        gConfig.forceRefresh()

        fullCfg = CFG()
        setup = gConfig.getValue("/DIRAC/Setup", "")
        setupList = gConfig.getSections("/DIRAC/Setups", [])
        if not setupList["OK"]:
            return S_ERROR("Could not get /DIRAC/Setups sections")
        setupList = setupList["Value"]
        if setup not in setupList:
            return S_ERROR("Setup %s is not in allowed list: %s" % (setup, ", ".join(setupList)))
        serviceSetups = gConfig.getOptionsDict("/DIRAC/Setups/%s" % setup)
        if not serviceSetups["OK"]:
            return S_ERROR("Could not get /DIRAC/Setups/%s options" % setup)
        serviceSetups = serviceSetups["Value"]  # dict
        for system, setup in serviceSetups.items():
            if self.systems and system not in self.systems:
                continue
            systemCfg = gConfigurationData.remoteCFG.getAsCFG("/Systems/%s/%s" % (system, setup))
            for section in systemCfg.listSections():
                if section not in ("Agents", "Services", "Executors"):
                    systemCfg.deleteKey(section)

            fullCfg.createNewSection("/%s" % system, contents=systemCfg)

        return S_OK(fullCfg)

    def _printDiff(self, entry, level=""):
        """Format the changes."""
        if len(entry) == 5:
            diffType, entryName, _value, changes, _comment = entry
        elif len(entry) == 4:
            diffType, entryName, _value, changes = entry

        fullPath = os.path.join(level, entryName)

        if diffType == "modSec":
            for change in changes:
                self._printDiff(change, fullPath)
        elif diffType == "modOpt":
            if self.showModified:
                LOG.notice("Changed option %r from %r" % (fullPath, changes))
        elif diffType == "delOpt":
            if self.showAdded:
                LOG.notice("Option %r does not exist in template" % fullPath)
        elif diffType == "delSec":
            if self.showAdded:
                LOG.notice("Section %r does not exist in template" % fullPath)
        elif diffType == "addSec":
            if self.showMissingSections:
                LOG.notice("Section %r not found in current configuration: %s" % (fullPath, pformat(changes)))
        elif diffType == "addOpt":
            if self.showMissingOptions:
                LOG.notice("Option %r not found in current configuration. Default value is %r" % (fullPath, changes))
        else:
            LOG.error("Unknown DiffType", "%s, %s, %s" % (diffType, fullPath, changes))

    def run(self):
        """Run configuration comparison."""
        self._setSwitches()
        self._check()
        return S_OK()


@Script()
def main():
    CheckConfig().run()


if __name__ == "__main__":
    main()



""" Test_RSS_Policy_JobEfficiencyPolicy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

import DIRAC.ResourceStatusSystem.Policy.JobEfficiencyPolicy as moduleTested

################################################################################


class JobEfficiencyPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """
        Setup
        """

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.JobEfficiencyPolicy

    def tearDown(self):
        """
        Tear down
        """

        del self.moduleTested
        del self.testClass


################################################################################


class JobEfficiencyPolicy_Success(JobEfficiencyPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        module = self.testClass()
        self.assertEqual("JobEfficiencyPolicy", module.__class__.__name__)

    def test_evaluate(self):
        """tests the method _evaluate"""

        module = self.testClass()

        res = module._evaluate({"OK": False, "Message": "Bo!"})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual("Bo!", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": None})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": []})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Completed": 0, "Done": 0, "Failed": 0}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("Not enough jobs to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Completed": 0, "Done": 0, "Failed": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("Not enough jobs to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Completed": 0, "Done": 0, "Failed": 10}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Banned", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency of 0.00", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Completed": 0, "Done": 8, "Failed": 2}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency of 0.80", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Completed": 10, "Done": 9, "Failed": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency of 0.95", res["Value"]["Reason"])


################################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(JobEfficiencyPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(JobEfficiencyPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""
   OpenNebula Endpoint base class implementation for the OpenNebula cloud service.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id"

import os
import base64
import ssl
from requests.auth import HTTPBasicAuth

try:
    from xmlrpc.client import ServerProxy
except ImportError:
    # python2 compat
    from xmlrpclib import ServerProxy

# DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Cloud.Endpoint import Endpoint
from DIRAC.Core.Utilities.File import makeGuid

DEBUG = False  # In case of debugging, switch


class OpenNebulaEndpoint(Endpoint):
    """OpenNebula implementation of the Cloud Endpoint interface"""

    def __init__(self, parameters=None):
        super(OpenNebulaEndpoint, self).__init__(parameters=parameters, bootstrapParameters=parameters)

        self.parameters["HostCert"] = self.parameters.get("HostCert")
        self.parameters["HostKey"] = self.parameters.get("HostKey")
        self.bootstrapParameters["HostCert"] = self.parameters["HostCert"]
        self.bootstrapParameters["HostKey"] = self.parameters["HostKey"]

        self.log = gLogger.getSubLogger("OpenNebulaEndpoint")

        self.valid = False
        self.vmType = self.parameters.get("VMType")
        self.site = self.parameters.get("Site")

        # Prepare the authentication request parameters
        self.session = None
        self.authArgs = {}
        self.user = self.parameters.get("User")
        self.password = self.parameters.get("Password")
        self.loginMode = False

        if self.user and self.password:
            # we have the login/password case
            self.authArgs["auth"] = HTTPBasicAuth(self.user, self.password)
            self.authArgs["verify"] = False
            self.loginMode = True
            self.oneauth = self.user + ":" + self.password
        else:
            # we have the user proxy case
            # TODO: test
            self.userProxy = os.environ.get("X509_USER_PROXY")
            self.userProxy = self.parameters.get("Proxy", self.userProxy)
            if self.userProxy is None:
                self.log.error("User proxy is not defined")
                self.valid = False
                return
            self.authArgs["cert"] = self.userProxy
            self.caPath = self.parameters.get("CAPath")
            self.authArgs["verify"] = self.caPath
            if self.parameters.get("Auth") == "voms":
                self.authArgs["data"] = '{"auth":{"voms": true}}'

        self.serviceUrl = self.parameters.get("EndpointUrl")

        result = self.initialize()
        if result["OK"]:
            self.log.debug("OpenNebulaEndpoint created and validated")
            self.valid = True

    def initialize(self):
        self.log.info("INITIALIZE")

        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE

        self.rpcproxy = ServerProxy(self.serviceUrl, context=ssl_ctx)

        return S_OK()

    def createInstances(self, vmsToSubmit):
        outputDict = {}
        message = ""
        self.log.debug("vmsToSubmit " + str(vmsToSubmit))
        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            createPublicIP = "ipPool" in self.parameters

            result = self.createInstance(instanceID, createPublicIP)
            if result["OK"]:
                nodeID, publicIP = result["Value"]
                self.log.debug("Created VM instance %s/%s with publicIP %s" % (nodeID, instanceID, publicIP))
                nodeDict = {}
                nodeDict["PublicIP"] = publicIP
                nodeDict["InstanceID"] = instanceID
                nodeDict["NumberOfProcessors"] = self.parameters.get("NumberOfProcessors", 1)
                outputDict[nodeID] = nodeDict
            else:
                message = result["Message"]
                break

        # We failed submission utterly
        if not outputDict:
            return S_ERROR("No VM submitted: %s" % message)

        return S_OK(outputDict)

    def createInstance(self, instanceID="", createPublicIP=True):
        """
        Creates VM in OpenNebula cloud.

        one.template.instantiate XML-RPC method is called with the following parameters
        This creates a VM instance for the given boot image
        and creates a context script, taken the given parameters.
        Successful creation returns instance VM

        Boots a new node on the OpenStack server defined by self.endpointConfig. The
        'personality' of the node is done by self.imageConfig. Both variables are
        defined on initialization phase.

        * template ID is obtained from Endpoint's `TemplateID` constructor parameter
        * VM name equals to `instanceID` method argument
        * onhold flag is obtained from Endpoint's `Onhold` constructor parameter
        * template string contains userdata scripts encoded in base64 and DNS
          servers adresses

        The node name has the following format:
        <bootImageName><contextMethod><time>

        :Parameters:
          **instanceID** - `string`
            name of a new VM
          **createPublicIP** - `bool`
            ignored

        It boots the node. If IPpool is defined on the imageConfiguration, a floating
        IP is created and assigned to the node.

        :return: S_OK( ( nodeID, publicIP ) ) | S_ERROR
        """

        # TODO: which templateId to use?
        templateId = int(self.parameters.get("TemplateID"))  # 1627 )
        onhold = self.parameters.get("Onhold", False)

        # TODO: remove this stuff (required by _createUserDataScript)
        self.parameters["VMUUID"] = instanceID
        self.parameters["VMType"] = self.parameters.get("CEType", "OpenNebula")
        result = self._createUserDataScript()
        if not result["OK"]:
            return result
        userData = str(result["Value"])

        # TODO: Should we have it as a parameter???
        template = """CONTEXT = [
NETWORK = "YES",
USER_DATA = "{}",
DNS = "8.8.8.8 8.8.4.4",
USERDATA_ENCODING = "base64"
]""".format(
            base64.b64encode(userData)
        )

        ret = self.rpcproxy.one.template.instantiate(self.oneauth, templateId, instanceID, onhold, template)

        # XXX: Sometimes it returns 4 values although documemntation says otherwise
        ok, payload, error = ret[:3]

        if ok:
            return S_OK((payload, None))

        return S_ERROR("one.template.instantiate failed with code: %s message: %s " % (error, payload))

    def getVMIDs(self):
        """Get all the VM IDs on the endpoint

        :return: list of VM ids
        """
        return S_ERROR("getVMIDs is not implemented")

    def getVMStatus(self, nodeID):
        """
        Get the status for a given node ID. libcloud translates the status into a digit
        from 0 to 4 using a many-to-one relation ( ACTIVE and RUNNING -> 0 ), which
        means we cannot undo that translation. It uses an intermediate states mapping
        dictionary, SITEMAP, which we use here inverted to return the status as a
        meaningful string. The five possible states are ( ordered from 0 to 4 ):
        RUNNING, REBOOTING, TERMINATED, PENDING & UNKNOWN.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )

        :return: S_OK( status ) | S_ERROR
        """
        return S_ERROR("getVMStatus is not implemented")

    def getVMNetworks(self):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        return S_ERROR("getVMNetworks is not implemented")

    def getVMNetworkInterface(self, network):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        return S_ERROR("getVMNetworkInterface is not implemented")

    def stopVM(self, nodeID, publicIP=""):
        """
        Terminates VM specified by nodeID by sending one.vm.action(..., 'delete', ..)
        request.

        :Parameters:
          **nodeID** - `string`
            OpenNebula VM id specifier
          **publicIP** - `string`
            ignored

        :return: S_OK | S_ERROR
        """
        self.log.info("DELETE ID: " + str(nodeID))

        ret = self.rpcproxy.one.vm.action(self.oneauth, "delete", int(nodeID))

        # XXX: Sometimes it returns 4 values although documemntation says otherwise
        ok = ret[0]
        payload = ret[1]
        error = ret[2]

        self.log.debug("Delete OK: " + str(ok))
        self.log.debug("Payload: " + str(payload))
        self.log.debug("Error: " + str(error))

        if ok:
            return S_OK(payload)

        return S_ERROR("one.vm.action failed with code: %s message: %s " % (error, payload))

    def assignFloatingIP(self, nodeID):
        """
        Given a node, assign a floating IP from the ipPool defined on the imageConfiguration
        on the CS.

        :Parameters:
          **node** - `libcloud.compute.base.Node`
            node object with the vm details

        :return: S_OK | S_ERROR
        """
        return S_ERROR("assignFloatingIP is not implemented")

    def getVMFloatingIP(self, publicIP):
        return S_ERROR("Not implemented")

    def deleteFloatingIP(self, nodeID):
        """
        Deletes a floating IP <public_ip> from the server.

        :param str publicIP: public IP to be deleted
        :param object node: node to which IP is attached
        :return: S_OK | S_ERROR
        """
        return S_ERROR("deleteFloatingIP is not implemnted")



########################################################################
# File: RequestValidator.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2012/09/18 07:55:16
########################################################################
""" :mod: RequestValidator

    ======================

    .. module: RequestValidator

    :synopsis: request validator

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    A general and simple request validator checking for required attributes and logic.
    It checks if required attributes are set/unset but not for their values.

    RequestValidator class implements the DIRACSingleton pattern, no global object is
    required to keep a single instance.

    If you need to extend this one with your own specific checks consider:

      * for adding Operation or Files required attributes use :any:`addReqAttrsCheck` function::

          RequestValidator().addReqAttrsCheck( "FooOperation", operationAttrs = [ "Bar", "Buzz"], filesAttrs = [ "LFN" ] )

      * for adding generic check define a new callable object ( function or functor ) which takes only one argument,
        say for functor::

          class MyValidator( RequestValidator ):

            @staticmethod
            def hasFoo( request ):
              if not request.Foo:
                return S_ERROR("Foo not set")
              return S_OK()

      * or function::

          def hasBar( request ):
            if not request.Bar:
              return S_ERROR("Bar not set")
            return S_OK()

    and add this one to the validators set by calling `RequestValidator().addValidator`, i.e.::

      RequestValidator().addValidator( MyValidator.hasFoo )
      RequestValidator().addValidator( hasFoo )

    Notice that all validators should always return S_ERROR/S_OK, no exceptions from that whatsoever!
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
# #
# @file RequestValidator.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/09/18 07:55:37
# @brief Definition of RequestValidator class.
# # import
import inspect
import six

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.Core.Security.Properties import FULL_DELEGATION, LIMITED_DELEGATION
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
from DIRAC.ConfigurationSystem.Client import PathFinder


########################################################################
@six.add_metaclass(DIRACSingleton)
class RequestValidator(object):
    """
    .. class:: RequestValidator

    This class validates newly created requests (before saving them in RequestDB) for
    required attributes.
    """

    # # dict with required attrs
    reqAttrs = {
        "ForwardDISET": {"Operation": ["Arguments"], "Files": []},
        "PutAndRegister": {"Operation": ["TargetSE"], "Files": ["LFN", "PFN"]},
        "ReplicateAndRegister": {"Operation": ["TargetSE"], "Files": ["LFN"]},
        "PhysicalRemoval": {"Operation": ["TargetSE"], "Files": ["LFN"]},
        "RemoveFile": {"Operation": [], "Files": ["LFN"]},
        "RemoveReplica": {"Operation": ["TargetSE"], "Files": ["LFN"]},
        "ReTransfer": {
            "Operation": ["TargetSE"],
            "Files": ["LFN", "PFN"],
        },
        "RegisterFile": {
            "Operation": [],
            "Files": ["LFN", "PFN", "ChecksumType", "Checksum", "GUID"],
        },
        "RegisterReplica": {
            "Operation": ["TargetSE"],
            "Files": ["LFN", "PFN"],
        },
    }

    # All the operationHandlers defined in the CS
    opHandlers = set()

    def __init__(self):
        """c'tor

        just setting validation order
        """
        self.validator = (
            self._hasRequestName,
            self._hasOwner,
            self._hasOperations,
            self._hasType,
            self._hasFiles,
            self._hasRequiredAttrs,
            self._hasChecksumAndChecksumType,
        )

        configPath = PathFinder.getAgentSection("RequestManagement/RequestExecutingAgent")

        # # operation handlers over here
        opHandlersPath = "%s/%s" % (configPath, "OperationHandlers")
        opHandlers = gConfig.getSections(opHandlersPath)
        if not opHandlers["OK"]:
            gLogger.error(opHandlers["Message"])
        else:
            self.opHandlers = set(opHandlers["Value"])

    @classmethod
    def addReqAttrsCheck(cls, operationType, operationAttrs=None, filesAttrs=None):
        """add required attributes of Operation of type :operationType:

        :param str operationType: Operation.Type
        :param operationAttrs: required Operation attributes
        :type operationAttrs: python:list
        :param filesAttrs: required Files attributes
        :type filesAttrs: python:list
        """
        toUpdate = {"Operation": operationAttrs if operationAttrs else [], "Files": filesAttrs if filesAttrs else []}
        if operationType not in cls.reqAttrs:
            cls.reqAttrs[operationType] = {"Operation": [], "Files": []}
        for key, attrList in cls.reqAttrs[operationType].items():
            cls.reqAttrs[operationType][key] = list(set(attrList + toUpdate[key]))

    @classmethod
    def addValidator(cls, fcnObj):
        """add `fcnObj` validator"""
        if not callable(fcnObj):
            return S_ERROR("supplied argument is not callable")
        args = inspect.getargspec(fcnObj).args
        if len(args) not in (1, 2):
            return S_ERROR("wrong number of arguments for supplied function object")
        cls.validator = cls.validator + tuple(
            fcnObj,
        )
        return S_OK()

    def validate(self, request):
        """validation of a given `request`

        :param ~Request.Request request: Request instance
        """
        for validator in self.validator:
            isValid = validator(request)
            if not isValid["OK"]:
                return isValid
        # # if we're here request is more or less valid
        return S_OK()

    @staticmethod
    def _hasDIRACSetup(request):
        """required attribute - DIRACSetup"""
        if not request.DIRACSetup:
            return S_ERROR("DIRACSetup not set")
        return S_OK()

    @staticmethod
    def _hasOwner(request):
        """required attributes OwnerDn and OwnerGroup"""
        if not request.OwnerDN:
            return S_ERROR("Request '%s' is missing OwnerDN value" % request.RequestName)
        if not request.OwnerGroup:
            return S_ERROR("Request '%s' is missing OwnerGroup value" % request.RequestName)
        return S_OK()

    @staticmethod
    def _hasRequestName(request):
        """required attribute: RequestName"""
        if not request.RequestName:
            return S_ERROR("RequestName not set")
        return S_OK()

    @staticmethod
    def _hasOperations(request):
        """at least one operation is in"""
        if not len(request):
            return S_ERROR("Operations not present in request '%s'" % request.RequestName)
        return S_OK()

    @staticmethod
    def _hasType(request):
        """operation type is set"""
        for operation in request:
            if not operation.Type:
                return S_ERROR(
                    "Operation #%d in request '%s' hasn't got Type set"
                    % (request.indexOf(operation), request.RequestName)
                )
        return S_OK()

    @classmethod
    def _hasFiles(cls, request):
        """check for files presence"""
        for operation in request:
            if operation.Type not in cls.reqAttrs:
                return S_OK()
            if cls.reqAttrs[operation.Type]["Files"] and not len(operation):
                return S_ERROR(
                    "Operation #%d of type '%s' hasn't got files to process."
                    % (request.indexOf(operation), operation.Type)
                )
            if not cls.reqAttrs[operation.Type]["Files"] and len(operation):
                return S_ERROR(
                    "Operation #%d of type '%s' has got files to process."
                    % (request.indexOf(operation), operation.Type)
                )
        return S_OK()

    @classmethod
    def _hasRequiredAttrs(cls, request):
        """check required attributes for operations and files"""
        for operation in request:
            if operation.Type in cls.reqAttrs:
                opAttrs = cls.reqAttrs[operation.Type]["Operation"]
                for opAttr in opAttrs:
                    if not getattr(operation, opAttr):
                        return S_ERROR(
                            "Operation #%d of type '%s' is missing %s attribute."
                            % (request.indexOf(operation), operation.Type, opAttr)
                        )
                fileAttrs = cls.reqAttrs[operation.Type]["Files"]
                for opFile in operation:
                    for fileAttr in fileAttrs:
                        if not getattr(opFile, fileAttr):
                            return S_ERROR(
                                "Operation #%d of type '%s' is missing %s attribute for file."
                                % (request.indexOf(operation), operation.Type, fileAttr)
                            )
        return S_OK()

    @classmethod
    def _hasChecksumAndChecksumType(cls, request):
        """Checksum and ChecksumType should be specified"""
        for operation in request:
            for opFile in operation:
                if any([opFile.Checksum, opFile.ChecksumType]) and not all([opFile.Checksum, opFile.ChecksumType]):
                    return S_ERROR(
                        "File in operation #%d is missing Checksum (%s) or ChecksumType (%s)"
                        % (request.indexOf(operation), opFile.Checksum, opFile.ChecksumType)
                    )
        return S_OK()

    def _hasExistingOperationTypes(self, request):
        """Check that there is a handler defined in the CS for each operation type"""
        requiredHandlers = set([op.Type for op in request])
        nonExistingHandlers = requiredHandlers - self.opHandlers

        if nonExistingHandlers:
            return S_ERROR(
                "The following operation type(s) have no handlers defined in the CS: %s" % nonExistingHandlers
            )

        return S_OK()

    @staticmethod
    def setAndCheckRequestOwner(request, remoteCredentials):
        """
        CAUTION: meant to be called on the server side.
                (does not make much sense otherwise)

        Sets the ownerDN and ownerGroup of the Request from
        the client's credentials.

        If they are already set, make sure the client is allowed to do so
        (FULL_DELEGATION or LIMITED_DELEGATION). This is the case of pilots or
        the RequestExecutingAgent

        :param request: the request to test
        :param remoteCredentials: credentials from the clients

        :returns: True if everything is fine, False otherwise
        """

        credDN = remoteCredentials["DN"]
        credGroup = remoteCredentials["group"]
        credProperties = remoteCredentials["properties"]

        # If the owner or the group was not set, we use the one of the credentials
        if not request.OwnerDN or not request.OwnerGroup:
            request.OwnerDN = credDN
            request.OwnerGroup = credGroup
            return True

        # From here onward, we expect the ownerDN/group to already have a value

        # If the credentials in the Request match those from the credentials, it's OK
        if request.OwnerDN == credDN and request.OwnerGroup == credGroup:
            return True

        # From here, something/someone is putting a request on behalf of someone else

        # Only allow this if the credentials have Full or Limited delegation properties

        if FULL_DELEGATION in credProperties or LIMITED_DELEGATION in credProperties:
            return True

        return False



""" Class that contains client access to the production DB handler. """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.ProductionSystem.Utilities.StateMachine import ProductionsStateMachine


@createClient("Production/ProductionManager")
class ProductionClient(Client):
    """Exposes the functionality available in the ProductionSystem/ProductionManagerHandler"""

    def __init__(self, **kwargs):
        """Simple constructor"""

        super(ProductionClient, self).__init__(**kwargs)
        self.setServer("Production/ProductionManager")
        self.prodDescription = {}
        self.stepCounter = 1

    # Method applying the Production System State machine

    def _applyProductionStatusStateMachine(self, prodID, status, force=False):
        """Performs a state machine check for productions when asked to change the status

        :param int prodID: the ProductionID on which the state machine check is performed
        :param str status: the proposed status which is checked to be valid
        :param bool force: a boolean. When force=True the proposed status is forced to pass the state machine check

        :return: S_OK with the new status or S_ERROR
        """
        res = self.getProductionParameters(prodID, "Status")
        if not res["OK"]:
            return res
        originalStatus = res["Value"]
        proposedStatus = status
        if force:
            return S_OK(proposedStatus)
        else:
            stateChange = ProductionsStateMachine(originalStatus).setState(proposedStatus)
            if not stateChange["OK"]:
                return S_OK(originalStatus)
            else:
                return S_OK(stateChange["Value"])

    # Methods contacting the ProductionManager Service

    def setProductionStatus(self, prodID, status):
        """Sets the status of a production

        :param int prodID: the ProductionID
        :param str status: the production status to be set to the prodID
        """
        rpcClient = self._getRPC()
        # Apply the production state machine
        res = self._applyProductionStatusStateMachine(prodID, status, force=False)
        if not res["OK"]:
            return res
        newStatus = res["Value"]
        if newStatus != status:
            return S_ERROR("Cannot change status")
        return rpcClient.setProductionStatus(prodID, status)

    def getProductions(self, condDict=None, older=None, newer=None, timeStamp=None, orderAttribute=None, limit=100):
        """Gets all the productions in the system, incrementally. "limit" here is just used to determine the offset."""

        rpcClient = self._getRPC()
        productions = []
        if condDict is None:
            condDict = {}
        if timeStamp is None:
            timeStamp = "CreationDate"
        # getting transformations - incrementally
        offsetToApply = 0
        while True:
            res = rpcClient.getProductions(condDict, older, newer, timeStamp, orderAttribute, limit, offsetToApply)
            if not res["OK"]:
                return res
            else:
                gLogger.verbose("Result for limit %d, offset %d: %d" % (limit, offsetToApply, len(res["Value"])))
                if res["Value"]:
                    productions = productions + res["Value"]
                    offsetToApply += limit
                if len(res["Value"]) < limit:
                    break
        return S_OK(productions)

    def getProductionTransformations(
        self, prodName, condDict=None, older=None, newer=None, timeStamp=None, orderAttribute=None, limit=10000
    ):
        """Gets all the production transformations for a production, incrementally.
            "limit" here is just used to determine the offset.

        :param str prodName: the production name
        :return: the list of the transformations associated to the production
        """

        rpcClient = self._getRPC()
        productionTransformations = []

        if condDict is None:
            condDict = {}
        if timeStamp is None:
            timeStamp = "CreationTime"
        # getting productionTransformations - incrementally
        offsetToApply = 0
        while True:
            res = rpcClient.getProductionTransformations(
                prodName, condDict, older, newer, timeStamp, orderAttribute, limit, offsetToApply
            )
            if not res["OK"]:
                return res
            else:
                gLogger.verbose("Result for limit %d, offset %d: %d" % (limit, offsetToApply, len(res["Value"])))
                if res["Value"]:
                    productionTransformations = productionTransformations + res["Value"]
                    offsetToApply += limit
                if len(res["Value"]) < limit:
                    break
        return S_OK(productionTransformations)

    def addProductionStep(self, prodStep):
        """Add a production step and update the production description

        :param object prodStep: the production step, i.e. a ProductionStep object describing the transformation
        """
        stepName = "Step" + str(self.stepCounter) + "_" + prodStep.Name
        self.stepCounter += 1
        prodStep.Name = stepName

        res = prodStep.getAsDict()
        if not res["OK"]:
            return res
        prodStepDict = res["Value"]
        rpcClient = self._getRPC()
        res = rpcClient.addProductionStep(prodStepDict)
        if not res["OK"]:
            return res
        stepID = res["Value"]
        self.prodDescription[prodStep.Name] = {"stepID": stepID, "parentStep": prodStepDict["parentStep"]}
        return S_OK()



""" Class defining a production step """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import json

from DIRAC import S_OK, S_ERROR


class ProductionStep(object):

    """Define the Production Step object"""

    def __init__(self, **kwargs):
        """Simple constructor"""
        # Default values for transformation step parameters
        self.Name = ""
        self.Description = "description"
        self.LongDescription = "longDescription"
        self.Type = "MCSimulation"
        self.Plugin = "Standard"
        self.AgentType = "Manual"
        self.FileMask = ""
        #########################################
        self.ParentStep = None
        self.Inputquery = None
        self.Outputquery = None
        self.GroupSize = 1
        self.Body = "body"

    def getAsDict(self):
        """It returns the Step description as a dictionary"""
        prodStepDict = {}
        prodStepDict["name"] = self.Name
        prodStepDict["parentStep"] = []
        # check the ParentStep format
        if self.ParentStep:
            if isinstance(self.ParentStep, list):
                prodStepDict["parentStep"] = []
                for parentStep in self.ParentStep:  # pylint: disable=not-an-iterable
                    if not parentStep.Name:
                        return S_ERROR("Parent Step does not exist")
                    prodStepDict["parentStep"].append(parentStep.Name)
            elif isinstance(self.ParentStep, ProductionStep):
                if not self.ParentStep.Name:
                    return S_ERROR("Parent Step does not exist")
                prodStepDict["parentStep"] = [self.ParentStep.Name]
            else:
                return S_ERROR("Invalid Parent Step")

        prodStepDict["description"] = self.Description
        prodStepDict["longDescription"] = self.LongDescription
        prodStepDict["stepType"] = self.Type
        prodStepDict["plugin"] = self.Plugin
        prodStepDict["agentType"] = self.AgentType
        prodStepDict["fileMask"] = self.FileMask
        # Optional fields
        prodStepDict["inputquery"] = json.dumps(self.Inputquery)
        prodStepDict["outputquery"] = json.dumps(self.Outputquery)
        prodStepDict["groupsize"] = self.GroupSize
        prodStepDict["body"] = json.dumps(self.Body)

        return S_OK(prodStepDict)



""" JobMonitoringHandler is the implementation of the JobMonitoring service
    in the DISET framework

    The following methods are available in the Service interface
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
from datetime import timedelta

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
import DIRAC.Core.Utilities.Time as Time
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.WorkloadManagementSystem.Service.JobPolicy import JobPolicy, RIGHT_GET_INFO

SUMMARY = []


class JobMonitoringHandlerMixin(object):
    @classmethod
    def initializeHandler(cls, svcInfoDict):
        """initialize DBs"""
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobDB", "JobDB")
            if not result["OK"]:
                return result
            cls.jobDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobLoggingDB", "JobLoggingDB")
            if not result["OK"]:
                return result
            cls.jobLoggingDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.TaskQueueDB", "TaskQueueDB")
            if not result["OK"]:
                return result
            cls.taskQueueDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.elasticJobParametersDB = None
        useESForJobParametersFlag = Operations().getValue("/Services/JobMonitoring/useESForJobParametersFlag", False)
        if useESForJobParametersFlag:
            try:
                result = ObjectLoader().loadObject(
                    "WorkloadManagementSystem.DB.ElasticJobParametersDB", "ElasticJobParametersDB"
                )
                if not result["OK"]:
                    return result
                cls.elasticJobParametersDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)

        cls.pilotManager = PilotManagerClient()
        return S_OK()

    @classmethod
    def parseSelectors(cls, selectDict=None):
        """Parse selectors before DB query

        :param dict selectDict: selectors

        :return: str, str, dict -- start/end date, selectors
        """
        selectDict = selectDict or {}

        # Get time period
        startDate = selectDict.get("FromDate", None)
        if startDate:
            del selectDict["FromDate"]
        # For backward compatibility
        if startDate is None:
            startDate = selectDict.get("LastUpdate", None)
            if startDate:
                del selectDict["LastUpdate"]
        endDate = selectDict.get("ToDate", None)
        if endDate:
            del selectDict["ToDate"]

        # Provide JobID bound to a specific PilotJobReference
        # There is no reason to have both PilotJobReference and JobID in selectDict
        # If that occurs, use the JobID instead of the PilotJobReference
        pilotJobRefs = selectDict.get("PilotJobReference")
        if pilotJobRefs:
            del selectDict["PilotJobReference"]
            if not selectDict.get("JobID"):
                for pilotJobRef in [pilotJobRefs] if isinstance(pilotJobRefs, six.string_types) else pilotJobRefs:
                    res = cls.pilotManager.getPilotInfo(pilotJobRef)
                    if res["OK"] and "Jobs" in res["Value"][pilotJobRef]:
                        selectDict["JobID"] = selectDict.get("JobID", [])
                        selectDict["JobID"].extend(res["Value"][pilotJobRef]["Jobs"])

        return startDate, endDate, selectDict

    @classmethod
    def getJobsAttributes(cls, *args, **kwargs):
        """Utility function for unpacking"""
        res = cls.jobDB.getJobsAttributes(*args, **kwargs)
        if not res["OK"]:
            return res
        return S_OK(strToIntDict(res["Value"]))

    ##############################################################################
    types_getApplicationStates = []

    @classmethod
    def export_getApplicationStates(cls, condDict=None, older=None, newer=None):
        """Return Distinct Values of ApplicationStatus job Attribute in WMS"""
        return cls.jobDB.getDistinctJobAttributes("ApplicationStatus", condDict, older, newer)

    ##############################################################################
    types_getJobTypes = []

    @classmethod
    def export_getJobTypes(cls, condDict=None, older=None, newer=None):
        """Return Distinct Values of JobType job Attribute in WMS"""
        return cls.jobDB.getDistinctJobAttributes("JobType", condDict, older, newer)

    ##############################################################################
    types_getOwners = []

    @classmethod
    def export_getOwners(cls, condDict=None, older=None, newer=None):
        """
        Return Distinct Values of Owner job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("Owner", condDict, older, newer)

    ##############################################################################
    types_getOwnerGroup = []

    @classmethod
    def export_getOwnerGroup(cls):
        """
        Return Distinct Values of OwnerGroup from the JobDB
        """
        return cls.jobDB.getDistinctJobAttributes("OwnerGroup")

    ##############################################################################
    types_getProductionIds = []

    @classmethod
    @deprecated("Unused")
    def export_getProductionIds(cls, condDict=None, older=None, newer=None):
        """
        Return Distinct Values of ProductionId job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("JobGroup", condDict, older, newer)

    ##############################################################################
    types_getJobGroups = []

    @classmethod
    def export_getJobGroups(cls, condDict=None, older=None, cutDate=None):
        """
        Return Distinct Values of ProductionId job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("JobGroup", condDict, older, newer=cutDate)

    ##############################################################################
    types_getSites = []

    @classmethod
    def export_getSites(cls, condDict=None, older=None, newer=None):
        """
        Return Distinct Values of Site job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("Site", condDict, older, newer)

    ##############################################################################
    types_getStates = []

    @classmethod
    def export_getStates(cls, condDict=None, older=None, newer=None):
        """
        Return Distinct Values of Status job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("Status", condDict, older, newer)

    ##############################################################################
    types_getMinorStates = []

    @classmethod
    def export_getMinorStates(cls, condDict=None, older=None, newer=None):
        """
        Return Distinct Values of Minor Status job Attribute in WMS
        """
        return cls.jobDB.getDistinctJobAttributes("MinorStatus", condDict, older, newer)

    ##############################################################################
    types_getJobs = []

    @classmethod
    def export_getJobs(cls, attrDict=None, cutDate=None):
        """
        Return list of JobIds matching the condition given in attrDict
        """
        # queryDict = {}

        # if attrDict:
        #  if type ( attrDict ) != dict:
        #    return S_ERROR( 'Argument must be of Dict Type' )
        #  for attribute in self.queryAttributes:
        #    # Only those Attribute in self.queryAttributes can be used
        #    if attrDict.has_key(attribute):
        #      queryDict[attribute] = attrDict[attribute]

        return cls.jobDB.selectJobs(attrDict, newer=cutDate)

    ##############################################################################
    types_getCounters = [list]

    @classmethod
    def export_getCounters(cls, attrList, attrDict=None, cutDate=""):
        """
        Retrieve list of distinct attributes values from attrList
        with attrDict as condition.
        For each set of distinct values, count number of occurences.
        Return a list. Each item is a list with 2 items, the list of distinct
        attribute values and the counter
        """

        _, _, attrDict = cls.parseSelectors(attrDict)
        return cls.jobDB.getCounters("Jobs", attrList, attrDict, newer=str(cutDate), timeStamp="LastUpdateTime")

    ##############################################################################
    types_getCurrentJobCounters = []

    @classmethod
    @deprecated("Unused")
    def export_getCurrentJobCounters(cls, attrDict=None):
        """Get job counters per Status with attrDict selection. Final statuses are given for
        the last day.
        """
        _, _, attrDict = cls.parseSelectors(attrDict)
        result = cls.jobDB.getCounters("Jobs", ["Status"], attrDict, timeStamp="LastUpdateTime")
        if not result["OK"]:
            return result
        last_update = Time.dateTime() - Time.day
        resultDay = cls.jobDB.getCounters("Jobs", ["Status"], attrDict, newer=last_update, timeStamp="LastUpdateTime")
        if not resultDay["OK"]:
            return resultDay

        resultDict = {}
        for statusDict, count in result["Value"]:
            status = statusDict["Status"]
            resultDict[status] = count
            if status in JobStatus.JOB_FINAL_STATES + JobStatus.JOB_REALLY_FINAL_STATES:
                resultDict[status] = 0
                for statusDayDict, ccount in resultDay["Value"]:
                    if status == statusDayDict["Status"]:
                        resultDict[status] = ccount
                    break

        return S_OK(resultDict)

    ##############################################################################
    types_getJobStatus = [int]

    @classmethod
    @deprecated("Use getJobsStatus")
    def export_getJobStatus(cls, jobID):

        return cls.jobDB.getJobAttribute(jobID, "Status")

    ##############################################################################
    types_getJobOwner = [int]

    @classmethod
    def export_getJobOwner(cls, jobID):

        return cls.jobDB.getJobAttribute(jobID, "Owner")

    ##############################################################################
    types_getJobSite = [int]

    @classmethod
    def export_getJobSite(cls, jobID):

        return cls.jobDB.getJobAttribute(jobID, "Site")

    ##############################################################################
    types_getJobJDL = [int, bool]

    @classmethod
    def export_getJobJDL(cls, jobID, original):

        return cls.jobDB.getJobJDL(jobID, original=original)

    ##############################################################################
    types_getJobLoggingInfo = [int]

    @classmethod
    def export_getJobLoggingInfo(cls, jobID):

        return cls.jobLoggingDB.getJobLoggingInfo(jobID)

    ##############################################################################
    types_getJobsParameters = [[six.string_types, int, list], list]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsParameters(cls, jobIDs, parameters):
        return cls.getJobsAttributes(jobIDs, parameters)

    ##############################################################################
    types_getJobsStates = [[six.string_types, int, list]]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsStates(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, ["Status", "MinorStatus", "ApplicationStatus"])

    ##############################################################################
    types_getJobsStatus = [[six.string_types, int, list]]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsStatus(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, ["Status"])

    ##############################################################################
    types_getJobsMinorStatus = [[six.string_types, int, list]]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsMinorStatus(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, ["MinorStatus"])

    ##############################################################################
    types_getJobsApplicationStatus = [[six.string_types, int, list]]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsApplicationStatus(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, ["ApplicationStatus"])

    ##############################################################################
    types_getJobsSites = [[six.string_types, int, list]]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobsSites(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, ["Site"])

    ##############################################################################
    types_getJobSummary = [int]

    @classmethod
    def export_getJobSummary(cls, jobID):
        return cls.jobDB.getJobAttributes(jobID, SUMMARY)

    ##############################################################################
    types_getJobPrimarySummary = [int]

    @classmethod
    @deprecated("Use getJobSummary")
    def export_getJobPrimarySummary(cls, jobID):
        return cls.jobDB.getJobAttributes(jobID, [])

    ##############################################################################
    types_getJobsSummary = [list]

    @classmethod
    def export_getJobsSummary(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, SUMMARY)

    ##############################################################################
    types_getJobPageSummaryWeb = [dict, list, int, int]

    def export_getJobPageSummaryWeb(self, selectDict, sortList, startItem, maxItems, selectJobs=True):
        """Get the summary of the job information for a given page in the
        job monitor in a generic format
        """

        resultDict = {}

        startDate, endDate, selectDict = self.parseSelectors(selectDict)

        # initialize jobPolicy
        credDict = self.getRemoteCredentials()
        ownerDN = credDict["DN"]
        ownerGroup = credDict["group"]
        operations = Operations(group=ownerGroup)
        globalJobsInfo = operations.getValue("/Services/JobMonitoring/GlobalJobsInfo", True)
        jobPolicy = JobPolicy(ownerDN, ownerGroup, globalJobsInfo)
        jobPolicy.jobDB = self.jobDB
        result = jobPolicy.getControlledUsers(RIGHT_GET_INFO)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("User and group combination has no job rights (%r, %r)" % (ownerDN, ownerGroup))
        if result["Value"] != "ALL":
            selectDict[("Owner", "OwnerGroup")] = result["Value"]

        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = sortList[0][0] + ":" + sortList[0][1]
        else:
            orderAttribute = None

        result = self.jobDB.getCounters(
            "Jobs", ["Status"], selectDict, newer=startDate, older=endDate, timeStamp="LastUpdateTime"
        )
        if not result["OK"]:
            return result

        statusDict = {}
        nJobs = 0
        for stDict, count in result["Value"]:
            nJobs += count
            statusDict[stDict["Status"]] = count

        resultDict["TotalRecords"] = nJobs
        if nJobs == 0:
            return S_OK(resultDict)

        resultDict["Extras"] = statusDict

        if selectJobs:
            iniJob = startItem
            if iniJob >= nJobs:
                return S_ERROR("Item number out of range")

            result = self.jobDB.selectJobs(
                selectDict, orderAttribute=orderAttribute, newer=startDate, older=endDate, limit=(maxItems, iniJob)
            )
            if not result["OK"]:
                return result

            summaryJobList = result["Value"]
            if not globalJobsInfo:
                validJobs, _invalidJobs, _nonauthJobs, _ownJobs = jobPolicy.evaluateJobRights(
                    summaryJobList, RIGHT_GET_INFO
                )
                summaryJobList = validJobs

            result = self.getJobsAttributes(summaryJobList, SUMMARY)
            if not result["OK"]:
                return result

            summaryDict = result["Value"]
            # If no jobs can be selected after the properties check
            if not summaryDict:
                return S_OK(resultDict)

            # Evaluate last sign of life time
            for jobID, jobDict in summaryDict.items():
                if "HeartBeatTime" not in jobDict or not jobDict["HeartBeatTime"] or jobDict["HeartBeatTime"] == "None":
                    jobDict["LastSignOfLife"] = jobDict["LastUpdateTime"]
                else:
                    lastTime = (
                        Time.fromString(jobDict["LastUpdateTime"])
                        if isinstance(jobDict["LastUpdateTime"], str)
                        else jobDict["LastUpdateTime"]
                    )
                    hbTime = (
                        Time.fromString(jobDict["HeartBeatTime"])
                        if isinstance(jobDict["HeartBeatTime"], str)
                        else jobDict["HeartBeatTime"]
                    )
                    # Not only Stalled jobs but also Failed jobs because Stalled
                    if (
                        (hbTime - lastTime) > timedelta(0)
                        or jobDict["Status"] == JobStatus.STALLED
                        or jobDict["MinorStatus"].startswith("Job stalled")
                        or jobDict["MinorStatus"].startswith("Stalling")
                    ):
                        jobDict["LastSignOfLife"] = jobDict["HeartBeatTime"]
                    else:
                        jobDict["LastSignOfLife"] = jobDict["LastUpdateTime"]

            # prepare the standard structure now
            key = list(summaryDict)[0]
            paramNames = list(summaryDict[key])

            records = []
            for jobID, jobDict in summaryDict.items():
                jParList = []
                for pname in paramNames:
                    jParList.append(jobDict[pname])
                records.append(jParList)

            resultDict["ParameterNames"] = paramNames
            resultDict["Records"] = records

        return S_OK(resultDict)

    ##############################################################################
    types_getJobStats = [six.string_types, dict]

    @classmethod
    def export_getJobStats(cls, attribute, selectDict):
        """Get job statistics distribution per attribute value with a given selection"""
        startDate, endDate, selectDict = cls.parseSelectors(selectDict)
        result = cls.jobDB.getCounters(
            "Jobs", [attribute], selectDict, newer=startDate, older=endDate, timeStamp="LastUpdateTime"
        )
        if not result["OK"]:
            return result
        resultDict = {}
        for cDict, count in result["Value"]:
            resultDict[cDict[attribute]] = count

        return S_OK(resultDict)

    ##############################################################################
    types_getJobsPrimarySummary = [list]

    @classmethod
    @ignoreEncodeWarning
    @deprecated("Use getJobsSummary")
    def export_getJobsPrimarySummary(cls, jobIDs):
        return cls.getJobsAttributes(jobIDs, [])

    ##############################################################################
    types_getJobParameter = [six.string_types + six.integer_types, six.string_types]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobParameter(cls, jobID, parName):
        """
        :param str/int jobID: one single Job ID
        :param str parName: one single parameter name
        """
        if cls.elasticJobParametersDB:
            res = cls.elasticJobParametersDB.getJobParameters(jobID, [parName])
            if not res["OK"]:
                return res
            if res["Value"].get(int(jobID)):
                return S_OK(res["Value"][int(jobID)])

        res = cls.jobDB.getJobParameters(jobID, [parName])
        if not res["OK"]:
            return res
        return S_OK(res["Value"].get(int(jobID), {}))

    ##############################################################################
    types_getJobOptParameters = [int]

    @classmethod
    def export_getJobOptParameters(cls, jobID):
        return cls.jobDB.getJobOptParameters(jobID)

    ##############################################################################
    types_getJobParameters = [six.string_types + six.integer_types + (list,)]

    @classmethod
    @ignoreEncodeWarning
    def export_getJobParameters(cls, jobIDs, parName=None):
        """
        :param str/int/list jobIDs: one single job ID or a list of them
        :param str parName: one single parameter name, a list or None (meaning all of them)
        """
        if cls.elasticJobParametersDB:
            if not isinstance(jobIDs, list):
                jobIDs = [jobIDs]
            parameters = {}
            for jobID in jobIDs:
                res = cls.elasticJobParametersDB.getJobParameters(jobID, parName)
                if not res["OK"]:
                    return res
                parameters.update(res["Value"])

            # Need anyway to get also from JobDB, for those jobs with parameters registered in MySQL or in both backends
            res = cls.jobDB.getJobParameters(jobIDs, parName)
            if not res["OK"]:
                return res
            parametersM = res["Value"]

            # and now combine
            final = dict(parametersM)
            # if job in JobDB, update with parameters from ES if any
            for jobID in final:
                final[jobID].update(parameters.get(jobID, {}))
            # if job in ES and not in JobDB, take ES
            for jobID in parameters:
                if jobID not in final:
                    final[jobID] = parameters[jobID]
            return S_OK(final)

        return cls.jobDB.getJobParameters(jobIDs, parName)

    ##############################################################################
    types_getAtticJobParameters = [list(six.integer_types)]

    @classmethod
    def export_getAtticJobParameters(cls, jobID, parameters=None, rescheduleCycle=-1):
        if not parameters:
            parameters = []
        return cls.jobDB.getAtticJobParameters(jobID, parameters, rescheduleCycle)

    ##############################################################################
    types_getJobAttributes = [int]

    @classmethod
    def export_getJobAttributes(cls, jobID, attrList=None):
        """
        :param int jobID: one single Job ID
        :param list attrList: optional list of attributes
        """

        return cls.jobDB.getJobAttributes(jobID, attrList=attrList)

    ##############################################################################
    types_getJobAttribute = [int, six.string_types]

    @classmethod
    def export_getJobAttribute(cls, jobID, attribute):
        """
        :param int jobID: one single Job ID
        :param str attribute: one single attribute name
        """

        return cls.jobDB.getJobAttribute(jobID, attribute)

    ##############################################################################
    types_getSiteSummary = []

    @classmethod
    def export_getSiteSummary(cls):
        return cls.jobDB.getSiteSummary()

    ##############################################################################
    types_getJobHeartBeatData = [int]

    @classmethod
    def export_getJobHeartBeatData(cls, jobID):
        return cls.jobDB.getHeartBeatData(jobID)

    ##############################################################################
    types_getInputData = [list(six.integer_types)]

    @classmethod
    def export_getInputData(cls, jobID):
        """Get input data for the specified jobs"""
        return cls.jobDB.getInputData(jobID)


##############################################################################


class JobMonitoringHandler(JobMonitoringHandlerMixin, RequestHandler):
    pass



########################################################################
# File :   Path.py
# Author : Ricardo Graciani
########################################################################
"""
Some Helper class to build CFG paths from tuples
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import six

cfgInstallSection = "LocalInstallation"
cfgResourceSection = "Resources"


def cfgPath(*args):
    """
    Basic method to make a path out of a tuple of string, any of them can be already a path
    """
    return os.path.normpath(os.path.join(*(str(k) for k in args)))


def cfgInstallPath(*args):
    """
    Path to Installation/Configuration Options
    """
    return cfgPath(cfgInstallSection, *args)


def cfgPathToList(arg):
    """
    Basic method to split a cfgPath in to a list of strings
    """
    listPath = []
    if not isinstance(arg, six.string_types):
        return listPath
    while arg.find("/") == 0:
        arg = arg[1:]
    return arg.split("/")
